# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/04_markdown.markdown.file.ipynb.

# %% ../../../nbs/04_markdown.markdown.file.ipynb 2
from __future__ import annotations
import copy
from enum import Enum
from itertools import product
from os import PathLike
import re
import string
from typing import Iterator, Union, Optional
import yaml

from trouver.helper import (
    find_regex_in_text 
)
from trouver.markdown.markdown.heading import (
    heading_level, heading_title
)
from trouver.markdown.obsidian.links import (
    ObsidianLink, 
    remove_links_from_text, EMBEDDED_PATTERN
)
from ..obsidian.vault import VaultNote, NoteDoesNotExistError, NotePathIsNotIdentifiedError
from trouver.markdown.obsidian.tags import (
    strip_auto_from_tag, tag_is_auto_tag
)

# %% auto 0
__all__ = ['find_front_matter_meta_in_markdown_text', 'dict_to_metadata_lines', 'replace_embedded_links_with_text',
           'MarkdownLineEnum', 'MarkdownFile']

# %% ../../../nbs/04_markdown.markdown.file.ipynb 5
def find_front_matter_meta_in_markdown_text(
        text: str
        ) -> Union[tuple[int], None]: # Each tuple is of the form `(a,b)` where `text[a:b]` is a markdown front matter. This list will be of length at most 1 because the front matter must come at the start of the markdown document.
    """
    Return ranges in the markdown text string where front matter meta occurs.
    
    `text` is assumed to start with the front matter meta.
    The front matter meta is of the following YAML format.
    
    ```markdown
    ---
    key: entry
    ---
    ```
    """
    listy = find_regex_in_text(text, pattern=r'---\n([\s\S])*?(?(1)\n|)---')
    return listy[0] if listy else None


# %% ../../../nbs/04_markdown.markdown.file.ipynb 11
def _enquote(str):
    return f'"{str}"'

# %% ../../../nbs/04_markdown.markdown.file.ipynb 13
def dict_to_metadata_lines(
        data: dict[str, Union[str, list[str]]] # The keys are str of the labels/names of the metadata. The values are the metadata, which are usually str or list.
        ) -> list[str]: # Each str entry is the line for the yaml frontmatter metadata of an Obsidian Markdown note.
    """
    Convert a dict to a list of str of yaml frontmatter metadata
    that Obsidian recognizes.

    This function is used in `MarkdownFile.replace_metadata`.
    
    **Returns**
    - list[str]
    """
    # lines = []
    # for key, value in data.items():
    #     value = '[' + ', '.join(value) + ']' if isinstance(value, list) else value
    #     lines.append(f'{key}: {value}')
    # return lines

    lines = []
    for key, value in data.items():
        if isinstance(value, list):
            _add_line_for_list(lines, key, value)
            continue
        lines.append(f'{key}: {value}')
    return lines

def _add_line_for_list(
        lines: list[str], key: str, list_value: list):
    """This is a helper function for `dict_to_metadata_lines`."""
    escaped_strings_in_list_value = [
        _enquote(single_value.replace('\\', '\\\\'))
        if isinstance(single_value, str) and ('\\' in single_value or '{' in single_value)
        else single_value
        for single_value in list_value]
    lines.append(f"{key}: [{', '.join(escaped_strings_in_list_value)}]")

# %% ../../../nbs/04_markdown.markdown.file.ipynb 20
def replace_embedded_links_with_text(
        text: str, vault: PathLike) -> str:
    """
    Return the text with all embedded links replaced with the
    text of the corresponding notes
    
    Assumes that the notes of the links exist in the vault and 
    have unique names in the vault. Note that embedded links are
    not always to notes (e.g. they can point to images), or even
    to existing notes. In such cases, the embedded link will be
    replaced with blank text.
    """
    
    embedded_links = find_regex_in_text(text, EMBEDDED_PATTERN)
    for start, end in reversed(embedded_links):
        link = text[start:end]
        link_object = ObsidianLink.from_text(link)
        try:
            vn = VaultNote(vault, name=link_object.file_name)
            replace = vn.text()
            #mf = MarkdownFile.from_vault_note(vn)
            #replace = str(mf)
        except NoteDoesNotExistError:
            replace = ''
        text = text[:start] + replace + text[end:]
    return text

# %% ../../../nbs/04_markdown.markdown.file.ipynb 22
class MarkdownLineEnum(Enum):
    # See https://www.markdownguide.org/basic-syntax/
    DEFAULT = 0
    HEADING = 1
    CODE_BLOCK = 2
    META = 3
    ORDERED_LIST = 4
    UNORDERED_LIST = 5
    BLOCKQUOTE = 6
    HORIZONTAL_RULE = 7
    COMMENT = 8
    BLANK_LINE = 9
    UNKNOWN = 10
    FOOTNOTE_DESCRIPTION = 11
    DISPLAY_LATEX_SINGLE = 12  # A single-line DISPLAY latex
    DISPLAY_LATEX_START = 13  # Start line of DISPLAY latex
    DISPLAY_LATEX_END = 14  # End line of DISPLAY latex
    DISPLAY_LATEX = 15  # All other DISPLAY latex lines

# %% ../../../nbs/04_markdown.markdown.file.ipynb 25
class MarkdownFile:
    r"""
    Parses and represents the contents of an Obsidian styled Markdown
    file.

    The Markdown file must be formatted in certain ways. In general, 
    text components of different types (see `MarkdownLineEnum`) must
    be on different lines - no text components of different types may
    occupy the same line. In particular,

    - Comments (surrounded by `%%`) must not be on the same line as
    non-comments.
    - Display math mode LaTeX (surrounded by `$$`) must not be on the same line
    as non-In line LaTeX.

    **Attributes**

    - text - str
    - parts - list[dict[str, Union[MarkdownLineEnum, str]]]
        - Represents the lines of the markdown file.
        Each dict has two keys, `'type'` and `'line'`, which respectively
        hold a `MarkdownLineEnum` and a `str` as values. Each value of
        `'line'` does not includes a new line character `\n` at the end
        by default.
    """
    
    def __init__(self, parts: list[dict[str, Union[MarkdownLineEnum, str]]]):
        # self.text = text
        self.parts = parts

    def __str__(self):
        return self.text_of_lines(0, len(self.parts))

    def text_of_lines(
            self,
            start: int,
            end: int
            ) -> str:
        """ Return the text of `self.parts[start:end]`. """
        return '\n'.join([
            line_dict['line'] for line_dict in self.parts[start:end]])

    def get_headings(
            self,
            levels: Union[int, Iterator[int], None] = None, # The levels of the headings to search for. Each int is between 1 and 6 inclusive, as each heading can be of levels 1 to 6. Defaults to `None`, in which case all heading-levels are searched.
            include_start: bool = True # If `True` and if this object contains text that is not under a heading (i.e. the text does not start with a heading), then include `-1` as a key with the empty str as value.
            ) -> list[str]: # Each str is the heading, including leading sharps `'#'`.
        """
        Return a list of heading titles in the markdown file.
        """
        line_dict = self.get_headings_by_line_number(
            levels, include_start=include_start)
        return list(line_dict.values())

    def get_headings_by_line_number(
            self,
            levels: Union[Iterator[int], int, None] = None, # The levels of the headings to search for. Each int is between 1 and 6 inclusive, as each heading can be of levels 1 to 6. If `None` then all heading-levels are searched.
            include_start: bool = True # If `True` and if this object contains text that is not under a heading (i.e. the text does not start with a heading), then include `-1` as a key with the empty str as value.
            ) -> dict[int, str]: # The keys are line numbers and each value is str is the heading string, including the leading sharps `'#'`, but without any leading or trailing whitespace characters.
        """
        Return a dict of heading titles in the markdown file.
        """
        if not levels:
            levels = []
        elif type(levels) is int:
            levels = (levels,)
        line_numbers = [i for i, line_dict in enumerate(self.parts)
                        if line_dict['type'] is MarkdownLineEnum.HEADING]
        extract = {line_number: self.parts[line_number] 
                   for line_number in line_numbers}
        headings = {line_number: line_dict['line'].strip()
                    for line_number, line_dict in extract.items()
                    if (not(levels) 
                        or heading_level(line_dict['line']) in levels)}
        if include_start and 0 not in headings:
            headings[-1] = ''
        return headings
    
    def get_headings_and_text(
            self,
            levels: Union[Iterator[int], int, None] = None, # The levels of the headings to search for. Each int is between 1 and 6 inclusive, as each heading can be of levels 1 to 6. If `None`, then all heading-levels are searched.
            include_start: bool = True # If `True` and if this object contains text that is not under a heading (i.e. the text does not start with a heading), then include `-1` as a key with the empty str as value.        
            ) -> dict[str, str]: # Each key is the entire str of the heading, including the leading sharps `'#'`, but not including leading or trailing whitespace characters Each value is the str under that heading until the next heading, including at trailing next line characters `\n`.  If `include_start` is `True`, then one of the keys is the empty str and the corresponding value is the start of the text that is not under any heading.
        # TODO I think there is a bug, e.g. when all headers are level 1 and 
        # level = 2, include_start = True, everything is treated like the start.
        """
        Return a list of headings and the text under each heading.
        
        The text under each heading does not include the text of
        subheadings.

        """
        headings_by_lines = self.get_headings_by_line_number(levels, include_start)
        line_numbers = [num for num in headings_by_lines]
        line_numbers.append(len(self.parts))
        line_numbers.sort()
        heading_dict = {}
        # previous_header_line_num = -1
        previous_header_line_num = -1 if include_start else line_numbers[0]
        for i, header_line_num in enumerate(line_numbers):
            lines = self.parts[previous_header_line_num+1:header_line_num]
            lines = [line_dict['line'] for line_dict in lines]
            heading = headings_by_lines.get(previous_header_line_num, '')
            previous_header_line_num = header_line_num
            heading_dict[heading] = '\n'.join(lines)
        return heading_dict
    
    def get_headings_tree(
            self
            ) -> dict[Union[str, int], Union[str, dict]]: # The keys are 1. line numbers or 2. the str `'title'`.  The values are dict or str (the blank str if root node) respectively. The dicts in themselves recursively represent trees and the str are headings, including the leading sharps.
        """Return a dict representing the tree of headings in the markdown file.

        **Returns**

        - dict[Union[str, int], Union[str, dict]]
            - The keys are 1. line numbers or 2. the str `'title'`.  The values
            are dict or str (the blank str if root node) respectively. The
            dicts in themselves recursively represent trees and the str are
            headings, including the leading sharps. In particular, the root
            level dict also has the blank string `''` associated to the key
            `'title'`.
        """
        headings_dict = self.get_headings_by_line_number(include_start=False)
        root_dict = {'title': ''}
        dict_stack = [root_dict]
        current_level = 0
        # Go through each heading, and figure out where it should
        # be added.
        for line_number, heading in headings_dict.items():
            new_dict = {'title': heading}
            while (heading_level(dict_stack[-1]['title'])
                   >= heading_level(heading)):
                dict_stack.pop()
            dict_stack[-1][line_number] = new_dict
            dict_stack.append(new_dict)
            current_level = heading_level(heading)
        return root_dict
                
    
    def get_line_number_of_heading(
            self,
            title: Union[str, None] = None, # Title of the heading. Does not include the leading sharps (`'#'`). If `None`, then return the line number of any heading after the specified line number.
            from_line: int = 0, # The line number to start searching for the heading with `title` from.
            levels: Union[Iterator[int], int, None] = None # The levels of the heading to search for. Each int is between 1 and 6 inclusive, as each heading can be of levels 1 to 6. If `None`, then all heading-levels are searched.
            ) -> int: # An index in `self.parts`. If no index/line number of the matching heading exists, then return -1.
        """
        Return the line number of the heading with the specified
        title after the specified line number.
        """
        if type(levels) is int:
            levels = (levels,)
        heading_satisfies = lambda x: (
            x['type'] is MarkdownLineEnum.HEADING
            and (levels is None or heading_level(x['line']) in levels)
            and (title is None or heading_title(x['line']) == title))
        return next((i for i in range(from_line, len(self.parts))
                     if heading_satisfies(self.parts[i])), -1)
    
    def get_line_numbers_under_heading(
            self,
            title: Union[str, None] = None, # Title of the heading. Does not include the leading sharps (`'#'`). If `None`, then return the line number of any heading after the specified line number.
            from_line: int = 0, # The line number to start searching for the heading with `title` from.
            levels: Union[Iterator[int], int, None] = None, # The levels of the heading to search for. Each int is between 1 and 6 inclusive, as each heading can be of levels 1 to 6. If `None`, then all heading-levels are searched.
            include_subheadings: bool = True # If `True`, then include the subheadings.
            ) -> Union[tuple[int], int]: # `(start, end)` where `self.parts[start:end]` represents the parts under the heading, including the start of the heading.  If the heading of the specified title does not exist, then returns -1.
        """
        Return the line numbers belonging to the heading.
        """
        # TODO: implement from_line, levels and then test
        start_line = self.get_line_number_of_heading(title)
        if start_line == -1:
            return -1
        level = heading_level(self.parts[start_line]['line'])
        if include_subheadings:
            levels_to_search = range(1, level+1)
        else:
            levels_to_search = None
        headings_and_lines = self.get_headings_by_line_number(
            levels=levels_to_search)
        lines_with_headings = list(headings_and_lines)
        lines_with_headings.sort()
        index_of_start = lines_with_headings.index(start_line)
        if len(lines_with_headings) - 1 >= index_of_start + 1:
            end_line = lines_with_headings[index_of_start + 1]
        else:
            end_line = len(self.parts)
        return (start_line, end_line)

    def write(
        self,
        vn: VaultNote, # Represents the file.
        mode: str = 'w' # The specific mode to write the file with.
        ) -> None:
        """
        Write to the file specified by a `VaultNote` object.

        If the file that the `VaultNote` object represents does not exist,
        then this method creates it.
        """
        if not vn.exists():
            vn.create()
        with open(vn.path(), mode, encoding='utf-8') as file:
            file.write(str(self))
            file.close()
            
    def insert_line(
            self,
            index: int, # The index at which to add `line_dict` into `self.parts`.
            line_dict: dict[str, Union[MarkdownLineEnum, str]] # See `self.parts`.
            ) -> None:
        """Add a line at the specified index/line number to `self.parts`."""
        self.parts.insert(index, line_dict)
        
    def remove_line(
            self,
            index: int = -1 # The index of the line to remove from `self.parts`.
            ) -> None:
        """Remove a line from `self.parts`."""
        del self.parts[index]
        
    def remove_lines(
            self,
            start: int, # The index of the first line to remove from `self.parts`.
            end: int # The end index to remove; the line of index `end` is not removed.
            ) -> None:
        """Remove lines from `self.parts`."""
        del self.parts[start:end]
        
    def pop_line(
            self,
            index: int = -1 # The index of the line to pop from `self.parts`.
            ) -> dict[str, Union[MarkdownLineEnum, str]]: # The popped line
        """Remove a line from `self.parts` and get its value."""
        return self.parts.pop(index)
        
    def add_line_to_end(
            self,
            line_dict: dict[str, Union[MarkdownLineEnum, str]] # See `self.parts`.
            ) -> None:
        """Add a line to the end of `self.parts`."""
        self.parts.append(line_dict)
        
    def add_blank_line_to_end(self) -> None:
        """Add a blank line to the end of `self.parts`."""
        # TODO: Should the blank line be '\n' or ''?
        self.add_line_to_end(
            {'type': MarkdownLineEnum.BLANK_LINE, 'line': '\n'})
        
    def add_line_in_section(
            self,
            title: str, # Title of the heading (without the leading sharps `'#'`)
            line_dict: dict[str, Union[MarkdownLineEnum, str]], # The line to add
            start: bool = True # If `True`, add to the start of the section. If `False`, add to the end of the section.
            ) -> None:
        # TODO start=False is not implemented. 
        # TODO Be able to tell what the inserted line's type is.
        # TODO This seems to work incorrectly if there are no lines
        # after a header.
        """Add a line in section specified by its title."""
        line_number = self.get_line_number_of_heading(title=title)
        self.insert_line(line_number + 1, line_dict)
     
    def remove_section(
            self,
            title: str # The title of the section to remove (without the starting `'#'`'s)
            ) -> None:
        """
        Remove the section with the specified title, including subsections,
        if the section exists.
        """
        section_line = self.get_line_number_of_heading(title=title)
        if section_line == -1:
            return
        level = heading_level(self.parts[section_line]['line'])
        big_level_lines = self.get_headings_by_line_number(range(1, level+1))
        line_numbers = [line_num for line_num in big_level_lines]
        line_numbers.sort()
        i = line_numbers.index(section_line) + 1
        next_section_line = (
            line_numbers[i] if i < len(line_numbers) else len(self.parts))
        self.remove_lines(section_line, next_section_line)
        
    
    def clear_section(
            self,
            title: str, # Title of the section (Without the leading sharps `'#'`)
            leave_blank_line: bool = True, # If `True`, leaves a blank line at the end of the section.
            clear_subsections: Optional[str] = None # `'clear'`, `'delete'`, or `None`. If `'clear'`, then just clears the contents of subsections, but does not affect the headers. If `'delete'`, then clears the contents of the subsections and deletes the headers. If `None`, then does not affect either. 
            ) -> None:
        # TODO: implement clear_subsections
        """
        Clear the section with the specified title, if it exists.
        
        Does not clear subsections.
        """
        section_line = self.get_line_number_of_heading(title=title)
        if section_line == -1:
            return
        next_section_line = self.get_line_number_of_heading(
            from_line=section_line+1)
        if next_section_line == -1:
            next_section_line = len(self.parts)
        self.parts = self.parts[:section_line+1] + self.parts[next_section_line:]
        if leave_blank_line:
            self.insert_line(
                section_line + 1,
                {'type': MarkdownLineEnum.BLANK_LINE, 'line': ''})

    def clear_all_sections(
            self,
            leave_blank_lines: bool = True
            # If True, leaves a blank line in each section
            ) -> None:
        """
        Clear all sections.
        
        Does not clear frontmatter metadata. Leaves all headers intact.
        """
        part_indices_to_remove = [
            i for (i, part) in enumerate(self.parts)
            if (part['type'] not in [MarkdownLineEnum.META,
                                     MarkdownLineEnum.HEADING])]
        for index_to_remove in reversed(part_indices_to_remove):
            self.remove_line(index_to_remove)
        
    def metadata(
            self
            ) -> Union[dict[str], None]: # The keys are `str` of the labels/names of the metadata. The values are the metadata, which are usually `str` or `list`. If there is not frontmatter YAML metadata, then this return value is `None`.
        # TODO: change the Exception to some kind of yaml format exception.
        """
        Return the frontmatter metadata as a dict.

        Writing the metadata read with this function can be faulty if
        the metadata attempts to escape characters that cannot be read by
        YAML; the escapes are not preserved upon writing.

        **Raises**

        - ValueError
            - If any exceptions are raised when reading (i.e. parsing or scanning
            the YAML metadata. In doing so, `str(self)` is printed. Moreover,
            the appropriate `yaml.parser.ParserError`, `yaml.scanner.ScannerError`,
            or `yaml.reader.ReaderError` is also raised.
        """
        metadata_parts = [part['line'] for part in self.parts 
                          if part['type'] == MarkdownLineEnum.META]
        try:
            return yaml.safe_load('\n'.join(metadata_parts[1:-1]))
        except (yaml.YAMLError) as e:
            raise ValueError(
                "There is invalid YAML formatting in a MarkdownFile object."
                " The following is its text:\n"
                f"{str(self)}\n\n") from e
        # try:
        #     return ruamel.yaml.load()
            
    def has_metadata(self) -> bool:
        """
        Return `True` if this `MarkdownFile` object has fronmatter
        YAML metadata.

        If the `MarkdownFile` object has any frontmatter YAML metadata, then
        it is expected to be at the very start; in particular, it must not
        be preceded by any whitespace characters.
        """
        return (self.parts 
                and self.parts[0]['type'] == MarkdownLineEnum.META)

    def add_metadata_section(
            self,
            check_exists: bool = True # If `True`, Check if there is already a metadata section at the beginning and do not add a metadata section if it exists.
            ) -> None:
        """Add a frontmatter YAML metadata at the very beginning."""
        if check_exists and self.has_metadata():
            return
        default_meta = [{'type': MarkdownLineEnum.META, 'line': '---\n'},
                        {'type': MarkdownLineEnum.META, 'line': '---\n'}]
        self.parts = default_meta + self.parts

    def metadata_lines(
            self
            ) -> tuple[int]: # The tuple consists of 2 ints, `a` and `b`, where `self.parts[a:b+1]` represent the metadata lines, including the `'---'` before and after.
        # TODO: change the index conventions of the output so that
        # [a:b] is the parts
        """
        Return the indices in `self.parts` which are metadata.
        
        Assumes that `self.parts` is nonempty. 

        If the MarkdownFile object has any frontmatter YAML metadata, then
        it is expected to be at the very start; in particular, it must not
        be preceded by any whitespace characters.
        
        **Returns**
         
        - tuple
            - The tuple consists of 2 ints, `a` and `b`, where `self.parts[a:b+1]`
            represent the metadata lines, including the `'---'` before and after.
        """
        for i, part in enumerate(self.parts):
            if part['type'] == MarkdownLineEnum.META:
                start = i
                break
        for j in range(i+1, len(self.parts)):
            if self.parts[j]['type'] != MarkdownLineEnum.META:
                end = j-1
                break
        return (start, end)
        
    def replace_metadata(
            self,
            new_metadata: dict[str] # The dictionary representing the new metadata.
        ) -> None:
        """Replace the frontmatter metadata of this MarkdownFile object."""
        lines = dict_to_metadata_lines(new_metadata)
        new_metadata_parts = [{'type': MarkdownLineEnum.META, 'line': line}
                              for line in lines]
        new_metadata_parts.insert(0, {'type': MarkdownLineEnum.META, 'line': '---'})
        new_metadata_parts.append({'type': MarkdownLineEnum.META, 'line': '---'})
        start, end = self.metadata_lines()
        self.parts = self.parts[:max(0,start-1)] + new_metadata_parts\
            + self.parts[end+1:]
    
    def remove_metadata(self) -> None:
        """Remove the frontmatter metadata of this MarkdownFile object."""
        self.parts = [part for part in self.parts 
                      if part['type'] != MarkdownLineEnum.META]
    
    def has_tag(
            self,
            tag: str # The tag. Does not start with the hashtag `'#'`.
            ) -> bool:
        """
        Return `True` if the Markdown file has the specified tag in its
        YAML frontmatter metadata.

        More specifically, return `True` if the `MarkdownFile` objeect

        1. has YAML frontmatter metadata,
        2. the metadata has a `'tags'` section,, and
        3. the `'tags'` section is a list with the specified tag.

        Note that `tag` should not start with the hashtag `#` charater.
        """
        if not self.has_metadata():
            return False
        metadata = self.metadata()
        return (bool(metadata) and 'tags' in metadata
                and tag in metadata['tags'])
    
    def add_tags(
            self,
            tags: list[str], # The str representing the tags. May or may not start with `'#'`, e.g. `'#_meta/definition'` or `'_meta/definition'`.
            skip_repeats: bool = True, # If `True`, then this MarkdownFile will just have unique tags; merges pre-existing repeated tags if necessary. Also, the order of the tags may be changed.
            skip_repeated_auto: bool = True # If `True`, then only add tags starting with '_auto/' if the corresponding non-auto tag does not exist, e.g.  '_auto/_meta/definition' is not added if the note already has '_meta/definition'.
            ) -> None:
        """
        Add tags to the frontmatter metadata.
        
        The order of the tags may be changed.
        """
        self.add_metadata_section(check_exists=True)
        metadata = self.metadata()
        if not metadata:
            metadata = {}
        if not 'tags' in metadata:
            metadata['tags'] = []

        tags = [tag[1:] if tag.startswith('#') else tag for tag in tags]
        if skip_repeated_auto:
            tags = [tag for tag in tags 
                    if (not tag_is_auto_tag(tag) 
                        or strip_auto_from_tag(
                            tag, with_hash_tag=False) not in metadata['tags'])]
        if skip_repeats:
            set_of_tags = set(metadata['tags']) | set(tags)
            metadata['tags'] = list(set_of_tags)
        else:
            metadata['tags'] += tags
        self.replace_metadata(metadata)
        
    def remove_tags(
            self,
            tags: list[str] # The str representing the tags. May or may not start with `'#'`, e.g. `'#_meta/definition'` or `'_meta/definition'`.
            ) -> None:
        """
        Remove specified tags from the frontmatter metadata, if
        the frontmatter metadata and the specified tags.

        If the `MarkdownFile` object does not have a frontmatter or
        if the frontmatter does not include a `tags` line, then
        the `MarkdownFile` object is not modified.
        
        Assumes that this MarkdownFile object has a frontmatter and
        that the frontmatter includes a tags line.
        
        Any repeated tags are either merged into one (if the tag is 
        not in `tags`) or are removed (if the tag is in `tags`).
        """
        tags = [tag[1:] if tag.startswith('#') else tag for tag in tags]
        metadata = self.metadata()
        if metadata is None or not 'tags' in metadata:
            return
        set_of_tags = set(metadata['tags'])
        set_of_tags -= set(tags)
        metadata['tags'] = list(set_of_tags)
        self.replace_metadata(metadata)
    
    def replace_auto_tags_with_regular_tags(
            self, exclude: list[str] = None
            # The tags whose `_auto/` tags should not be converted. The str should not start with `'#'` and should not start with `'_auto/'`.
            ) -> None:
        """
        Replace tags in the frontmatter metadata starting with `_auto/`
        with tags without the `_auto/`.
        """
        if not exclude:
            exclude = []
        metadata = self.metadata()
        no_auto_tags = [strip_auto_from_tag(tag, with_hash_tag=False)
                         for tag in metadata['tags']]
        new_tags = [
            no_auto if (tag_is_auto_tag(old) and no_auto not in exclude)
            else old for (no_auto, old) in zip(no_auto_tags, metadata['tags'])
        ]
        metadata['tags'] = new_tags
        self.replace_metadata(metadata)
    
    def remove_in_line_tags(self) -> None:
        """Remove lines starting with in line tags."""
        part_indices_to_remove = []
        for i, part in enumerate(self.parts):
            if (part['type'] == MarkdownLineEnum.DEFAULT
                    and part['line'].strip().startswith('#')):
                part_indices_to_remove.append(i)
        for i in reversed(part_indices_to_remove):
            self.remove_line(i)
            
    def replace_links_with_display_text(
            self,
            remove_embedded_note_links: bool = False # If `True`, remove links to embedded notes as well. If `False`, does not modify embedded notes.`
            ) -> None:
        """Remove nonembedded links and replaces them with their display text.
        """
        for part in self.parts:
            if part['type'] == MarkdownLineEnum.META:
                continue
            part['line'] = remove_links_from_text(
                part['line'],
                remove_embedded_note_links=remove_embedded_note_links)
    
    def parts_of_id(
            self,
            par_id: str # Must begin with `'^'`.
            ) -> Union[tuple[int], None]: # `(start,end)` where `self.parts[start:end]` consists of the lines of the specified id. If the specified id does not exist for the note, then `None` is returned.
        """
        Return the indices of the lines within the Markdown file
        belonging to the specified text id.

        This id can be used as an anchor for a link in Obsidian. For example,
        `[[note#^65809f]]` is a link to a note named `note` to the text with id
        `65809f`. Such a text is marked with a trailing `^65809f`.
        """
        pattern = re.compile(fr'(?:^|\s){re.escape(par_id)}\b(?:\s*?)$')
        for i, part in enumerate(reversed(self.parts)):
            match = re.search(pattern, part['line'])
            if match:
                break

        end_of_text = len(self.parts) - 1 - i
        i = end_of_text
        has_any = _remove_text_id(self.parts[i]['line']).strip() != ''
        cont = True
        while i > 0 and cont:
            cont, has_any = self._include_previous_line_as_id_text(i, has_any)
            if cont:
                i -= 1
        return i, end_of_text+1
        # self.parts[i:end_of_text+1]

    def _include_previous_line_as_id_text(self, i: int, has_any: bool):
        """Used in `text_of_id`

        **Parameters**
        - i - int
            - Greater than 0.
        - has_any - bool
            - `True` if any text beyond the id has been gathered.
            `False` otherwise.

        **Return**
        - bool, bool
            - The first bool is `True` if the previous line should be
            included in the text of the id. The second bool is `True` if
            any content any been included in the search.
        """

        if self.parts[i]['type'] == MarkdownLineEnum.HEADING:
            return False, True
        elif self.parts[i-1]['type'] == MarkdownLineEnum.HEADING:
            return False, has_any
        elif (self.parts[i]['type'] == MarkdownLineEnum.DISPLAY_LATEX
                and self.parts[i-1]['type'] == MarkdownLineEnum.DISPLAY_LATEX):
            return True, True
        elif self.parts[i-1]['line'].strip() == '':
            if has_any:
                return False, True
            else:
                return True, False
        return True, True

    def replace_embedded_links_with_text(
            self,
            vault: PathLike,
            recursive: bool = True, # If `True`, then recursively replaces embedded links in the text of the embedded links.
            remove_paragraph_id: bool = True # If `True`, then removes the paragraph id's in the text of the embedded links. Leaves the paragraph id's of the origianl text in tact.
            ) -> None:
        """
        Remove embedded links and replaces them with their underlying text,
        as found in notes in the vault.
        
        Assumes that the embedded links do not loop infinitely.

        For embedded links to notes that do not exist in the vault,
        the embedded links are replaced with blank str.

        No new entries are added to `self.parts` even if the embedded links
        have multiple lines.
        """
        # TODO: implement error raise upon infinite loop.
        for part in self.parts:
            part['line'] = self._replace_embedded_links_one_line(
                part['line'], vault, recursive, remove_paragraph_id)
    
    def _replace_embedded_links_one_line(
            self, text: str, vault: PathLike, recursive: bool,
            remove_paragraph_id: bool) -> str:
        # TODO: test what happens when `link_note` is actually an image.
        # TODO: deal with possibility that a vaultnote gets an image
        # file name passed.
        """Used in `replace_embedded_links_with_text`"""
        embedded_links = find_regex_in_text(text, EMBEDDED_PATTERN)
        for start, end in reversed(embedded_links):
            link_object = ObsidianLink.from_text(text[start:end])
            if link_object.file_name:
                try:
                    link_note = VaultNote(vault, name=link_object.file_name)
                    link_file = MarkdownFile.from_vault_note(link_note)
                except (NoteDoesNotExistError, NotePathIsNotIdentifiedError):
                    text = text[:start] + text[end:]
                    continue
                # if not link_note.exists():
                #     text = text[:start] + text[end:]
                #     continue
            else:
                link_file = self
            if link_object.anchor == 0:
                replace = str(link_file)
            elif link_object.anchor.startswith('^'):
                replace = self._text_of_embedded_link_of_id(
                    link_file, link_object.anchor, vault, recursive, remove_paragraph_id)
            elif link_object.anchor:
                replace = self._text_of_embedded_link_of_section(
                    link_file, link_object.anchor, vault, recursive, remove_paragraph_id)
            else:
                replace = str(link_file)
            text = text[:start] + replace + text[end:]
        return text
            # try:
            #     vn = VaultNote(vault, name=link_object.file_name)
            #     mf = MarkdownFile.from_vault_note(vn)
            # except NoteDoesNotExistError:
            #     replace = ''
            # text = text[:start] + replace + text[end:]

    def _text_of_embedded_link_of_id(
            self, link_file: MarkdownFile, par_id: str, vault: PathLike,
            recursive: bool, remove_paragraph_id: bool) -> str:
        """Used in `_replace_embedded_links_one_line"""
        start, end = link_file.parts_of_id(par_id)
        embedded_text = self._text_of_lines_of_embedded_links(
            link_file, vault, start, end, recursive,
            remove_paragraph_id)
        if remove_paragraph_id:
            embedded_text = _remove_text_id(embedded_text)
        return embedded_text

    def _text_of_embedded_link_of_section(
            self, link_file: MarkdownFile, heading: str, vault: PathLike,
            recursive: bool, remove_paragraph_id: bool) -> str:
        """Used in `_replace_embedded_links_one_line"""
        start, end = link_file.get_line_numbers_under_heading(
            title=heading, include_subheadings=True)
        return self._text_of_lines_of_embedded_links(
            link_file, vault, start, end, recursive,
            remove_paragraph_id)

    def _text_of_lines_of_embedded_links(
            self, link_file: MarkdownFile, vault: PathLike,
            start: int, end: int, recursive: bool,
            remove_paragraph_id: bool) -> str:
        """Used in `_text_of_embedded_link_of_id` and 
        `_text_of_embedded_link_of_section`
        """
        if not recursive:
            return link_file.text_of_lines(start, end)
        new_mf = MarkdownFile(
            copy.deepcopy(link_file.parts[start:end]))
        new_mf.replace_embedded_links_with_text(
            vault, recursive, remove_paragraph_id)
        return str(new_mf)

    def remove_footnotes_to_embedded_links(
            self,
            remove_footnote_mentions: bool = True # If `True`, removes the mentions to the footnote to the embedded links in the text.
            ) -> None:
        """
        Remove footnotes to embedded links.
        
        These are footnotes whose only content are embedded links, e.g.
        `[^1]: ![[embedded_note]]`
        """
        footnote_parts_to_remove = [
            (i, part) for i, part in enumerate(self.parts)
            if part['type'] == MarkdownLineEnum.FOOTNOTE_DESCRIPTION
            and re.fullmatch(
                fr'\[\^\w+\]: ?{EMBEDDED_PATTERN}', part['line'].strip())]
        for i, _ in reversed(footnote_parts_to_remove):
            self.remove_line(i)
        if not remove_footnote_mentions:
            return
        footnote_labels_to_remove = [
            re.fullmatch(
                fr'\[\^(\w+)\]: ?{EMBEDDED_PATTERN}', part['line'].strip()).group(1)
            for _, part in footnote_parts_to_remove]
        for part, label in product(self.parts, footnote_labels_to_remove):
            part['line'] = part['line'].replace(f'[^{label}]', '')
            
    def remove_headers(self) -> None:
        """Remove all headers."""
        heading_lines = self.get_headings_by_line_number(include_start=False)
        heading_lines = [line for line in heading_lines]
        heading_lines.sort()
        for line in reversed(heading_lines):
            self.remove_line(line)
        
    def remove_double_blank_lines(self) -> None:
        """Remove blank lines so that there are no consecutive blank lines"""
        parts_to_remove = []
        for i, part in enumerate(self.parts):
            last_blank = self.parts[i-1]['type'] == MarkdownLineEnum.BLANK_LINE
            if (part['type'] == MarkdownLineEnum.BLANK_LINE and last_blank):
                parts_to_remove.append(i)
        for i in reversed(parts_to_remove):
            self.remove_line(i)
    
    def merge_display_math_mode(self) -> None:
        """Merge chunks of display_math_mode latex lines into single lines"""
        # TODO: test
        i = 0
        ils = MarkdownLineEnum.DISPLAY_LATEX_START
        while i < len(self.parts):
            if self.parts[i]['type'] == ils:
                self._merge_one_display_math_mode_latex_chunk(i)
            i += 1
        
    def _merge_one_display_math_mode_latex_chunk(self, start: int) -> None:
        """Merge one chunk of display math mode latex lines in self.parts.
        """
        j = start + 1 
        ile = MarkdownLineEnum.DISPLAY_LATEX_END
        while (j < len(self.parts)
               and self.parts[j]['type'] is not ile):
            j += 1
        end = j + 1
        lines = [part['line'] for part in self.parts[start:end]]
        merged = ' '.join(lines)
        self.remove_lines(start, end)
        self.insert_line(
            start, {'type': MarkdownLineEnum.DISPLAY_LATEX_SINGLE,
                    'line': merged})

    def merge_display_math_mode_into_preceding_text(
            self,
            separator: str = '\n' # The str with which to join the latex lines into the text lines. Note that the display math mode latex lines are not joined with this str.
            ) -> None:
        """
        Merge chunks of display math mode latex lines into single lines and merge
        those single lines into preceding text lines.
        """
        self.merge_display_math_mode()
        i = 0
        ils = MarkdownLineEnum.DISPLAY_LATEX_SINGLE
        while i < len(self.parts):
            if self.parts[i]['type'] == ils:
                i = self._merge_latex_into_text(i, separator)
            i += 1

    def _merge_latex_into_text(self, index: int, separator: str) -> int:
        # TODO: test
        """Used in `merge_display_math_mode_into_preceding_text`"""
        j = index-1
        if j == -1:
            return index
        while j >= 0 and self.parts[j]['line'].strip() == '':
            j -= 1
        merged = separator.join([self.parts[i]['line'] for i in range(j, index+1)])
        self.remove_lines(start=j+1, end=index+1)
        self.parts[j]['line'] = merged
        return j

    @classmethod
    def from_file(
            cls,
            file_path: PathLike
            ) -> MarkdownFile:
        """
        Return a `MarkdownFile` object from a specified file.
        
        **Raises**
        - FileNotFoundError
            - If `file_path` points to a file which does not exist.
        """
        with open(file_path, 'r', encoding='utf-8') as file:
            text = file.read()
        lines = text.split('\n')
        return cls.from_list(lines)
            
    @classmethod
    def from_list(
            cls,
            list_of_lines: list[str]
            ) -> MarkdownFile:
        """
        Return a `MarkdownFile` object from a list of lines.
        
        This may not work correctly if the markdown text is not
        sufficiently well-formatted. These formattings include:
        - comments must start the line with `'%%'`.
        - comments must end with `'%%'` followed by whitespaces
          and nothing else.
        - indents should be done with tabs?
        """
        parts = []
        parts.extend(cls._look_at_start_of_file(list_of_lines))
        for line in list_of_lines:
            parts.append(cls._line_dict(line, parts))
        return cls(parts)
    
    @classmethod
    def from_vault_note(cls, vn: VaultNote) -> MarkdownFile:
        """
        Return a `MarkddownFile` object from a `VaultNote` object.
        
        **Raises**
        - FileNotFoundError
            - If `vn` represents a note file which does not exist.
        """
        return cls.from_file(vn.path())
    
    @classmethod
    def from_string(cls, text: str) -> MarkdownFile:
        """
        Return a `MarkdownFile` object from a str.
        """
        return cls.from_list(text.splitlines(keepends=False))
    
    def copy(self, deep: bool) -> MarkdownFile:
        # TODO: test
        list_to_copy_with = self.parts
        if deep:
            list_to_copy_with = copy.deepcopy(list_to_copy_with)
        return self.__class__(list_to_copy_with)

    @classmethod
    def _look_at_start_of_file(cls, list_of_lines: list[str])\
            -> list[dict[str, Union[MarkdownLineEnum, str]]]:
        """
        Inspect start of file for blank lines and Front matter Meta.
        
        Only blank lines may preceed front matter in markdown.
        May remove leading entries in `list_of_lines`.
        
        **Parameters**
        - list_of_lines - list of str

        **Returns**
        - list[dict[str, Union[MarkdownLineEnum, str]]]
            - Each dict has two keys, 'type' and 'line', which respectively
            hold a `MarkdownLineEnum` and a str as values.
            
        **Notes**
        - First tries to find the opening of the front matter given by
        `'---'`, then tries to find the closing given by `'---'`.
        """
        # Find opening of front matter
        i = 0
        parts = []
        while list_of_lines and list_of_lines[0].isspace():
            parts.append({'type': MarkdownLineEnum.BLANK_LINE,
                      'line': list_of_lines[0]})
            del list_of_lines[0]
        
        if not list_of_lines or list_of_lines[i].strip() != '---':
            return parts

        # Find close of front matter
        found_close_of_front_matter = False
        for i, line in enumerate(list_of_lines):
            if i == 0:
                continue
            if line.strip() == '---':
                found_close_of_front_matter = True
                break
        if not found_close_of_front_matter:
            return parts
        for j in range(0,i+1):
            parts.append({'type': MarkdownLineEnum.META,
                          'line': list_of_lines[0]})
            del list_of_lines[0]
        return parts
        
    @classmethod
    def _line_dict(cls, line: str, parts: list[dict])\
            -> dict[str, Union[MarkdownLineEnum, str]]:
        """
        Return a dictionary giving some information about `line`.
        
        **Parameters**
        - `line` - str
        - parts - dict[str, Union[MarkdownLineEnum, str]]
            - Each dict is the dictionary of a previous line.
        
        **Returns**
        - dict[str, Union[MarkdownLineEnum, str]]
            - Has a key `'type'`, which holds a MarkdownLineEnum
            as its value.
            - Has a key `'line'`, which holds `line`.
            - If the type is MarkdownLineEnum.CODE_BLOCK
        """
        line_dict = {'line': line}
        # Check previous line to see 
        if parts and parts[-1]['type'] is MarkdownLineEnum.COMMENT:
            if not parts[-1]['line'].rstrip().endswith('%%'):
                line_dict['type'] = MarkdownLineEnum.COMMENT
                return line_dict
        elif line.startswith('%%'):
            line_dict['type'] = MarkdownLineEnum.COMMENT
        if parts and parts[-1]['type'] is MarkdownLineEnum.CODE_BLOCK:
            if parts[-1]['line'].strip() != '```':
                line_dict['type'] = MarkdownLineEnum.CODE_BLOCK
                return line_dict
        elif line.strip().startswith('```'):
            line_dict['type'] = MarkdownLineEnum.COMMENT
        ###
        if parts and parts[-1]['type'] in [
                MarkdownLineEnum.DISPLAY_LATEX,
                MarkdownLineEnum.DISPLAY_LATEX_START]:
            if _line_end_in_line_latex(line) or '$$' in line:
                line_dict['type'] = MarkdownLineEnum.DISPLAY_LATEX_END
            else:
                line_dict['type'] = MarkdownLineEnum.DISPLAY_LATEX
            return line_dict


        if line.lstrip() == '':
            line_dict['type'] = MarkdownLineEnum.BLANK_LINE
        elif _line_start_and_end_in_line_latex(line):
            line_dict['type'] = MarkdownLineEnum.DISPLAY_LATEX_SINGLE
        elif _line_start_in_line_latex(line):
            line_dict['type'] = MarkdownLineEnum.DISPLAY_LATEX_START
        elif line.lstrip().startswith('%%'):
            line_dict['type'] = MarkdownLineEnum.COMMENT
        elif re.match(r'^#{1,6} ', line):
            line_dict['type'] = MarkdownLineEnum.HEADING
        elif line == '---':
            line_dict['type'] = MarkdownLineEnum.HORIZONTAL_RULE
        elif line.lstrip().startswith(('-', '*', '+')):
            line_dict['type'] = MarkdownLineEnum.UNORDERED_LIST
        elif re.match(r'^\d+\.', line.lstrip()):
            line_dict['type'] = MarkdownLineEnum.ORDERED_LIST
        elif line.lstrip().startswith('>'):
            line_dict['type'] = MarkdownLineEnum.BLOCKQUOTE
        elif re.match(r'^\[\^.+?\]:', line):
            line_dict['type'] = MarkdownLineEnum.FOOTNOTE_DESCRIPTION
        else:
            stripped_line = line.strip()
            if stripped_line.startswith(('---', '>', '# ', '##')):
                line_dict['type'] = MarkdownLineEnum.UNKNOWN
            else:
                line_dict['type'] = MarkdownLineEnum.DEFAULT
        return line_dict


def _remove_text_id(line: str) -> str:
    """"""
    # ID either starts at the start of line or is preceded by a whitespace
    # ID consists of \w\d.
    # ID ends with word boundary and might be followed by whitespaces.
    return re.subn(r'((?:^|\s)\^[\w\d]+)\b(?:\s*?)$', r'', line)[0]

def _line_start_in_line_latex(line: str) -> bool:
    """Return `True` if the line starts an display math mode latex string.
    
    This is not fully accurate outside of context - e.g. 
    It will return `True` on `$$`, regardless of whether this
    actually starts or ends the display math mode latex.
    """
    return line.lstrip(string.whitespace + '*').startswith('$$') 

def _line_end_in_line_latex(line: str) -> bool:
    stripped = _remove_text_id(line)
    stripped = stripped.rstrip(string.whitespace + '*')
    return stripped.endswith('$$')

def _line_start_and_end_in_line_latex(line: str) -> bool:
    stripped = _remove_text_id(line)
    stripped = stripped.strip(string.whitespace + '*')
    return bool(re.match(r'\$\$.*\$\$', stripped))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/16_latex.convert.ipynb.

# %% auto 0
__all__ = ['DEFAULT_NUMBERED_ENVIRONMENTS', 'DEFAULT_ENVIRONMENTS_TO_NOT_DIVIDE_ALONG', 'UNTITLED_SECTION_TITLE',
           'COMMON_SECTION_TITLES', 'remove_comments', 'divide_preamble', 'NoDocumentNodeError', 'find_document_node',
           'environment_names_used', 'numbered_newtheorems_counters_in_preamble', 'numberwithins_in_preamble',
           'display_names_of_environments', 'get_node_from_simple_text', 'swap_numbers_invoked', 'divide_latex_text',
           'section_and_subsection_titles_from_latex_parts', 'custom_commands', 'regex_pattern_detecting_command',
           'replace_command_in_text', 'replace_commands_in_text', 'replace_commands_in_latex_document',
           'adjust_common_syntax_to_markdown', 'setup_reference_from_latex_parts']

# %% ../../nbs/16_latex.convert.ipynb 3
from collections import OrderedDict
from itertools import product
import os
from os import PathLike
from pathlib import Path
import re
from typing import Optional, Union

from pathvalidate import sanitize_filename
from pylatexenc import latexwalker, latex2text
from pylatexenc.latexwalker import (
    LatexWalker, LatexEnvironmentNode, get_default_latex_context_db,
    LatexNode, LatexSpecialsNode, LatexMathNode, LatexMacroNode, LatexCharsNode,
    LatexGroupNode, LatexCommentNode
)
from pylatexenc.latex2text import (
    MacroTextSpec, EnvironmentTextSpec)
from pylatexenc.macrospec import (
    MacroSpec, LatexContextDb, EnvironmentSpec
)
import regex

from trouver.helper import (
    find_regex_in_text, dict_with_keys_topologically_sorted,
    containing_string_priority, replace_string_by_indices, text_from_file
)
from trouver.markdown.markdown.file import (
    MarkdownFile, MarkdownLineEnum
)
from ..markdown.obsidian.links import ObsidianLink

from ..markdown.obsidian.vault import VaultNote
from trouver.markdown.obsidian.personal.index_notes import (
    correspond_headings_with_folder, convert_title_to_folder_name
)
from ..markdown.obsidian.personal.reference import setup_folder_for_new_reference
from ..markdown.obsidian.vault import VaultNote
import warnings

# %% ../../nbs/16_latex.convert.ipynb 4
DEFAULT_NUMBERED_ENVIRONMENTS = ['theorem', 'corollary', 'lemma', 'proposition',
                                 'definition', 'conjecture', 'remark', 'example',
                                 'question']

# %% ../../nbs/16_latex.convert.ipynb 14
def remove_comments(text: str) -> str:
    # Find all occurrences of the comment pattern %[^\n]*
    return re.sub(r"%[^\n]*", "", text)

# %% ../../nbs/16_latex.convert.ipynb 21
def divide_preamble(
        text: str, # LaTeX document
        document_environment_name: str = "document"
        ) -> tuple[str, str]:
    """Divide the preamble from the rest of a LaTeX document.
    """
    begin_environment_str = rf'\begin{{{document_environment_name}}}'
    pattern = re.compile(re.escape(begin_environment_str))
    match = re.search(pattern, text) 
    start_match, end_match = match.span()
    return text[:start_match], text[start_match:]

    

# %% ../../nbs/16_latex.convert.ipynb 24
class NoDocumentNodeError(Exception):
    """Exception raised when a LatexEnvironmentNode corresponding to the document 
    environment is expected in a LaTeX string, but no such node exists.
    
    **Attributes**
    - text - str
        - The text in which the document environment is not found.
    """
    
    def __init__(self, text):
        self.text = text
        super().__init__(
            f"The following text does not contain a document environment:\n{text}")



# %% ../../nbs/16_latex.convert.ipynb 25
def find_document_node(
        text: str, # LaTeX str
        document_environment_name: str = "document" # The name of the document environment.
        ) -> LatexEnvironmentNode:
    """Find the `LatexNode` object for the main document in `text`.
    
    **Raises**
    - NoDocumentNodeError
        - If document environment node is not detected.
    """
    w = LatexWalker(text)
    nodelist, _, _ = w.get_latex_nodes(pos=0)
    for node in nodelist:
        if node.isNodeType(LatexEnvironmentNode)\
                and node.environmentname == document_environment_name:
            return node
    raise NoDocumentNodeError(text)

# %% ../../nbs/16_latex.convert.ipynb 36
def environment_names_used(
        text: str # LaTeX document
        ) -> set[str]: # The set of all environment names used in the main document.
    """Return the set of all environment names used in the main document
    of the latex code.
    """
    document_node = find_document_node(text)
    return {node.environmentname for node in document_node.nodelist
            if node.isNodeType(LatexEnvironmentNode)}        

# %% ../../nbs/16_latex.convert.ipynb 45
def _search_counters_by_pattern(
        preamble: str,
        newtheorem_regex: re.Pattern, # This is supposed to be a regex that detects and captures parameters of `\newtheorem` commands.
        counter_group: int # This depends on which `newtheorem_regex` is used, and is either 3 or 4. 
        ) -> dict[str, str]: # The 
    """
    Capture the newly defined theorem-like environment names as well as the
    counters that they belong to
    
    This is a helper function for `numbered_newtheorems_counters_in_preamble`.
    
    """
    counters = {}
    for match in newtheorem_regex.finditer(preamble):
        env_name = match.group(1)
        counter = match.group(counter_group)
        # If no counter was specified, use the environment name as the counter
        if counter is None:
            counter = env_name
        counters[env_name] = counter
    return counters

# %% ../../nbs/16_latex.convert.ipynb 48
def numbered_newtheorems_counters_in_preamble(
        document: str # The LaTeX document
        ) -> dict[str, tuple[str, Union[str, None]]]: # The keys are the command names of the environments.  The value a key is a tuple `(<counter>, <reset_by_counter>)`, where `<counter>`` is the counter that the environment belongs to, which can be custom defined or predefined in LaTeX, and `<reset_by_counter>` is a counter whose incrementation resets the # counter of the environment, if available. 
    r"""Return the dict specifying the numbered `\newtheorem` command invocations

    Assumes that

    - invocations of the `\newtheorem` command are exclusively in the
    preamble of the LaTeX document.
    - theorem-like environments are defined using the `\newtheorem` command.
    - no environments of the same name are defined twice.

    This function does not take into account `numberwithins` being used.

    This function uses two separate regex patterns, one to detect the invocations
    of `\newtheorem` in which the optional parameter is the second parameter and
    one to detect those in which the optional parameter is the third parameter.


    """
    preamble, _ = divide_preamble(document)
    preamble = remove_comments(preamble)
    # TODO: maybe use the `regex` package instead of `re` with a recursive
    # balanced-curly braces detecting regex.

    # matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{proposition}[theorem]{Proposition}`
    # does not match `\newtheorem{theorem}{Theorem}[Section]`
    second_parameter_pattern = re.compile(
        # In this case, the optional parameter (if any) should not follow the newtheorem.
        r'\\newtheorem\s*\{\s*(\w+)\s*\}\s*(\[\s*(\w+)\s*\])?\s*\{\s*(.*)\s*\}(?!\s*\[\s*(\w+)\s*\])')
    # matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{theorem}{Theorem}[Section]`,
    # does not match `\newtheorem{proposition}[theorem]{Proposition}`
    third_parameter_pattern = re.compile(
        r'\\newtheorem\s*\{\s*(\w+)\s*\}\s*\{\s*(.*)\s*\}\s*(\[\s*(\w+)\s*\])?')

    # TODO: return a dict whose values are tuples.
    second_results = _search_counters_by_pattern(preamble, second_parameter_pattern, 3)
    third_results = _search_counters_by_pattern(preamble, third_parameter_pattern, 4)
    to_return = {}
    for environment_name, counter in second_results.items():
        to_return[environment_name] = (counter, None)
    for environment_name, reset_counter in third_results.items():
        if environment_name in to_return:
            continue
        to_return[environment_name] = (environment_name, reset_counter)
    return to_return        

# %% ../../nbs/16_latex.convert.ipynb 59
def numberwithins_in_preamble(
        document: str # The LaTeX document
    ) -> dict[str, str]: # The keys are the first arguments of `numberwithin` invocations and the values ar ethe second arguments of `numberwithin` invocations.
    r"""Return the dict describing `numberwithin` commands invoked
    in the preamble of `document`."""
    preamble, _ = divide_preamble(document)
    preamble = remove_comments(preamble)
    pattern = regex.compile(r'\\numberwithin\s*\{\s*(\w+)\s*\}\s*\{\s*(.*)\s*\}')
    numberwithins = {}

    for match in pattern.finditer(preamble):
        environment_to_number = match.group(1)
        environment_to_count = match.group(2)
        numberwithins[environment_to_number] = environment_to_count

    

    return numberwithins

# %% ../../nbs/16_latex.convert.ipynb 63
def display_names_of_environments(
        document: str # The LaTeX document
        ) -> dict[str, str]:  
    r"""Return the dict specifying the display names for each theorem-like
    environment.

    This function uses two separate regex patterns, one to detect the invocations
    of `\newtheorem`
    in which the optional parameter is the second parameter and one to detect
    those in which the optional parameter is the third parameter.

    Assumes that
    - invocations of the `\newtheorem` command are exclusively in the
    preamble of the LaTeX document.
    - theorem-like environments are defined using the `\newtheorem` command.
    - no environments of the same name are defined twice.

    """
    preamble, _ = divide_preamble(document)
    # matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{proposition}[theorem]{Proposition}`
    # does not match `\newtheorem{theorem}{Theorem}[Section]`
    second_parameter_pattern = re.compile(
        # In this case, the optional parameter (if any) should not follow the newtheorem.
        r'\\newtheorem\*?\s*\{\s*(\w+\*?)\s*\}\s*(\[\s*(\w+)\s*\])?\s*\{\s*(.*)\s*\}(?!\s*\[\s*(\w+)\s*\])')
    # matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{theorem}{Theorem}[Section]`,
    # does not match `\newtheorem{proposition}[theorem]{Proposition}`
    third_parameter_pattern = re.compile(
        r'\\newtheorem\*?\s*\{\s*(\w+\*?)\s*\}\s*\{\s*(.*)\s*\}\s*(\[\s*(\w+)\s*\])?')
    second_results = _search_display_names_by_pattern(preamble, second_parameter_pattern, 4)
    third_results = _search_display_names_by_pattern(preamble, third_parameter_pattern, 2)
    return second_results | third_results
    

def _search_display_names_by_pattern(
        preamble: str,
        newtheorem_regex: re.Pattern,
        display_name_group: int # This depends on which `newtheorem_regex` is used, and is either 3 or 4. 
        ) -> dict[str, str]:
    """
    Capture the newly defined theorem-like environment names as well as the
    counters that they belong to"""
    display_names = {}
    for match in newtheorem_regex.finditer(preamble):
        env_name = match.group(1)
        display_name = match.group(display_name_group)
        display_names[env_name] = display_name
    return display_names

# %% ../../nbs/16_latex.convert.ipynb 67
def _setup_counters(
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`
        ) -> dict[str, int]:
    r"""
    Return a dict whose keys are of counters in the LaTeX document and whose
    values are all `0`. These key-value pairs are used to keep track of
    the numberings of `parts`.

    One special key is the key of the empty string `''`, which counters the
    parts which do not get a numbering, i.e. for most text that lie outside
    of (numbered) environments

    """
    # TODO: replace enumerated environments with markdown enumerated lists
    # and itemizes with markdown bulleted lists

    # cf. https://www.overleaf.com/learn/latex/Counters#Default_counters_in_LaTeX
    predefined_counters = [
        'part', # Incremented each time the `\part` command is used. It is not reset automatically and casn only be reset by the user
        'chapter', # Incremeneted each time the `\chapter` command is used.
        'section', # Incremented whenever a new `\section` command is encountered
        'subsection', # Incremented whenever a new `\subsection` command is encountered, reset whenever a new `\section` command is encountered
        'subsubsection', # Incremented whenever a new `\subsubsection` command is encounted, reset whenever a new `\subsection` or `\section` command is encountered
        'paragraph', # Incremeneted whenever a new paragraph is started. Reset whenever a new `\subsubsection`, `\subsection`, or `\section` command is encounted
        'subparagraph', # Incremented each time the `\subparagraph` command is used and reset at the beginning of a new
        'page', # Incremented each time a new page is started in the document
        'equation', # Incremeneted whenever the `\begin{equation}` environment is used. 
        'figure', # Incremented whenever a new `figure` environment is encountered
        'table', # Incremeneted whenever a new `taable` environment is encountered`
        'footnote', 
        'mpfootnote',
        'enumi',
        'enumii',
        'enumiii',
        'enumiv']

    counters = {counter: 0 for _, (counter, reset_counter) in numbertheorem_counters.items()}
    for counter in predefined_counters:
        counters[counter] = 0

    counters[''] = 0
    return counters

# %% ../../nbs/16_latex.convert.ipynb 69
def _setup_numberwithins(
        explicit_numberwithins: dict[str, str],
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`.
        ) -> dict[str, str]: # The keys are counters and the values are all counters that the key is immediately numbered within.
    """
    Extracts information of counters that are reset when other counters are
    incremented.

    This is a helper function of `_setup_all_numberwithins` as well as
    `divide_latex_text`.
    """
    builtin_numberwithins = {
        'subsection': 'section',
        'subsubsection': 'subsection',
        'paragraph': 'subsubsection',
        'subparagraph': 'paragraph',
        'enumii': 'enumi',
        'enumiii': 'enumii',
        'enumiv': 'enumiii',
        'part': 'chapter',
        'appendix': 'chapter'
    }
    numberwithins = explicit_numberwithins | builtin_numberwithins

    for environmentname, (counter, reset_by_counter) in numbertheorem_counters.items():
        if reset_by_counter is None:
            continue
        numberwithins[environmentname] = reset_by_counter
    return numberwithins

    

def _setup_all_numberwithins(
        explicit_numberwithins: dict[str, str],
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`.
        ) -> dict[str, list[str]]: # The keys are counters and the values are all counters that the key is numbered within.
    """
    This is a helper function of `divide_latex_text`.
    """
    numberwithins = _setup_numberwithins(explicit_numberwithins, numbertheorem_counters)
    all_counters = set()
    for key, value in numberwithins.items():
        all_counters.add(key)
        all_counters.add(value)
    all_numbered_withins = {counter: [] for counter in all_counters}
    for counter_1, counter_2 in product(all_counters, all_counters):
        if _is_numberedwithin(counter_1, counter_2, numberwithins):
            all_numbered_withins[counter_1].append(counter_2)
    return all_numbered_withins


def _is_numberedwithin(
        counter_1, counter_2, numberwithins: dict[str, str]
        ) -> bool:
    """Return `True` if `counter_1` is numbered within `counter_2""" 
    if counter_1 not in numberwithins:
        return False
    elif numberwithins[counter_1] == counter_2:
        return True
    return _is_numberedwithin(
        numberwithins[counter_1], counter_2, numberwithins)


# %% ../../nbs/16_latex.convert.ipynb 71
def _unnumbered_environments(
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`
        display_names: dict[str, str]) -> set[str]:
    r"""Return the set of unnumbered theorem-like environments defined by
    `\newtheorem`.

    This is a helper function of `divide_latex_text`.
    """
    return {environment for environment in display_names
            if environment not in numbertheorem_counters}

    

# %% ../../nbs/16_latex.convert.ipynb 73
def _section_title(
        text: str
        ) -> tuple[bool, str]: # The bool is `True` if the section/subsection is numbered (i.e. is `section` or `subsection` as opposed to `section*` or `subsection*`). The `str` is the title of the section or subsection
    """Return the title of a section or subsection from a latex str
    and whether or not the section/subsection is numbered"""

    # Note that the `section` command has the optional argument `toc-title` which appears
    # in the table of contents, cf.
    # http://latexref.xyz/_005csection.html
    pattern = regex.compile(
        r'\\(?:section|subsection)\s*(?:\[.*\])?(\*)?\s*'
        r'\{((?>[^{}]+|\{(?2)\})*)\}',
        regex.MULTILINE
    )
    regex_search = regex.search(pattern, text)
    is_numbered = regex_search.group(1) is None
    title = regex_search.group(2)
    return is_numbered, title


# %% ../../nbs/16_latex.convert.ipynb 75
def _is_section_node(node):
    return (node.isNodeType(LatexMacroNode)
            and node.macroname == 'section')

def _is_subsection_node(node):
    return (node.isNodeType(LatexMacroNode)
            and node.macroname == 'subsection')

def _is_environment_node(node):
    return node.isNodeType(LatexEnvironmentNode)

# %% ../../nbs/16_latex.convert.ipynb 77
def _is_numbered(
        node: LatexNode,
        numbertheorem_counters: dict[str, str]
        ) -> bool:
    if _is_section_node(node) or _is_subsection_node(node):
        is_numbered, _ = _section_title(node.latex_verbatim())
        return is_numbered
    elif _is_environment_node(node):
        return node.environmentname in numbertheorem_counters
    else:
        return False

# %% ../../nbs/16_latex.convert.ipynb 79
def _change_counters(
        node,
        counters,
        numbertheorem_counters: dict[str, str],
        all_numberwithins: dict[str, list[str]]
        ):
    # identify which counter to change
    # TODO
    # Take into consideration unnumbered non-environment node
    # Take into consideration unnumbered environment node
    if _is_environment_node(node):
        if node.environmentname in numbertheorem_counters:
           counter = numbertheorem_counters[node.environmentname][0]
        else:
            counter = None
    elif _is_section_node(node):
        counter = 'section'
    elif _is_subsection_node(node):
        counter = 'subsection'
    else:
        counter = None

    # Section counters seem to only reset subsection counters
    # When the section is numbered, etc., cf. `numbering_example_4...`
    # and `numbering_example_5...` in `nbs\_tests\latex_examples`
    is_numbered = _is_numbered(node, numbertheorem_counters)
    # e.g. `\numberwithin{equation}{section}`` means that `equation` is
    # numbered within `section`, i.e. `equation` is reset whenever
    # `section` is incremeneted

    # if counter is None and not _is_environment_node(node):
    #     counters[''] += 1 
    #     return

    if is_numbered:
        counters[counter] += 1
    for numbered_counter, within_counter in all_numberwithins.items():
        if counter is not None and counter in within_counter:
            counters[numbered_counter] = 0



# %% ../../nbs/16_latex.convert.ipynb 80
def get_node_from_simple_text(
        text: str) -> LatexNode:
    """Return the (first) `LatexNode` object from a str."""
    w = LatexWalker(text)
    nodelist, _, _ = w.get_latex_nodes(pos=0)
    return nodelist[0]

# %% ../../nbs/16_latex.convert.ipynb 83
def _node_numbering(
        node: LatexNode,
        numbertheorem_counters: dict[str, str],
        numberwithins: dict[str, str],
        counters: dict[str, int]
        ) -> str: # Just the numbering of the node, no "section/subsection" or displayname
    if _is_section_node(node):
        counter = 'section'
    elif _is_subsection_node(node):
        counter = 'subsection'
    elif _is_environment_node(node):
        counter = numbertheorem_counters[node.environmentname][0]
    return _numbering_helper('', counter, numberwithins, counters)


def _numbering_helper(
        trailing_numbering: str,
        counter: str,
        numberwithins: dict[str, str],
        counters: dict[str, int]
        ) -> str:
    """Recurisve helper function to `_node_numbering`."""
    if counter not in numberwithins and counter not in counters:
        return trailing_numbering
    if counter not in numberwithins and counter in counters and trailing_numbering:
        return f'{counters[counter]}.{trailing_numbering}'
    if counter not in numberwithins and counter in counters and not trailing_numbering:
        return f'{counters[counter]}'

    parent_counter = numberwithins[counter]
    current_count = counters[counter]
    if not trailing_numbering:
        to_pass_to_trailing_numbering = str(current_count)
    else:
        to_pass_to_trailing_numbering = f'{current_count}.{trailing_numbering}'

    return _numbering_helper(
        to_pass_to_trailing_numbering,
        parent_counter,
        numberwithins,
        counters)
    

# %% ../../nbs/16_latex.convert.ipynb 85
def _title(
        node: LatexNode,
        numbertheorem_counters: dict[str, str],
        numberwithins: dict[str, str], # An output of _setup_numberwithins
        all_numberwithins: dict[str, list[str]], # An output of all_numberwithins
        display_names: dict[str, str],
        counters: dict[str, int],
        swap_numbers: bool):
    """Return the title of a node based on the count in
    `counters`"""
    numbered = _is_numbered(node, numbertheorem_counters)
    if _is_section_node(node) and numbered:
        _, title = _section_title(node.latex_verbatim())
        return f"{counters['section']}. {title}"
    if _is_section_node(node) and not numbered:
        _, title = _section_title(node.latex_verbatim())
        return title 

    if _is_subsection_node(node) and numbered:
        _, title = _section_title(node.latex_verbatim())
        return f"{counters['section']}.{counters['subsection']}. {title}"
    if _is_subsection_node(node) and not numbered:
        _, title = _section_title(node.latex_verbatim())
        return title

    if _is_environment_node(node):
        return _title_for_environment_node(
            node, numbertheorem_counters, numberwithins,
            display_names, counters, swap_numbers)


def _title_for_environment_node(
        node: LatexNode,
        numbertheorem_counters: dict[str, str],
        numberwithins: dict[str, list[str]],
        display_names: dict[str, str],
        counters: dict[str, int],
        swap_numbers: bool):
    """Return the title of an environment node.
    If the node is not that of an theorem-like environment, then 
    
    """
    numbered = _is_numbered(node, numbertheorem_counters)
    # TODO: see what happens when environments are numbered within
    # sections vs. subsections
    if not numbered:
        numbering = None
    else:
        numbering = _node_numbering(
            node, numbertheorem_counters, numberwithins, counters)
    
    environment = node.environmentname
    if environment in display_names:
        display_name = display_names[environment]
    else:
        display_name = environment
    if not numbered:
        return display_name
    elif swap_numbers:
        return f'{numbering}. {display_name}.'
    else:
        return f'{display_name} {numbering}.'
        

# %% ../../nbs/16_latex.convert.ipynb 87
def swap_numbers_invoked(
        preamble: str
        ) -> bool: # 
    """Returns `True` if `\swapnumbers` is in the preamble.

    Assume that a mention of `\swapnumbers` is an actual invocation.
    """
    preamble = remove_comments(preamble)
    return '\swapnumbers' in preamble

# %% ../../nbs/16_latex.convert.ipynb 89
def _node_warrants_own_part(
        node, environments_to_not_divide_along: list[str],
        accumulation: str, parts: list[tuple[str, str]]) -> bool:
    """Return `True` if `node` warrants making a new part to be added in `parts`.

    This is a helper function for `_process_node`. When this function returns
    `true`, the `accumulation` should be considered for appending to `parts`
    and the node should also be appended to `parts
    and 
    """
    if _is_section_node(node) or _is_subsection_node(node):
        return True
    elif not _is_environment_node(node):
        return False
    # Is environment node from here and below.
    if len(parts) == 0 and accumulation.strip() == '':
        return True
    # elif len(parts) == 0:
    #     return node.environmentname not in environments_to_not_divide_along
    # previous_node = get_node_from_simple_text(parts[-1][1])
    # if (accumulation.strip() == ''
    #         and (_is_section_node(previous_node)
    #              or _is_subsection_node(previous_node))):
    #     return True
    return node.environmentname not in environments_to_not_divide_along

# %% ../../nbs/16_latex.convert.ipynb 91
def _node_is_proof_immediately_following_a_theorem_like_environment(
        node, accumulation, parts, display_names) -> bool:
    """Return `True` if `node` is that of a proof environment that immediately
    follows a theorem-like environment.
    
    This is a helper function for `_process_node`.
    """
    if not _is_environment_node(node):
        return False
    if not node.environmentname == 'proof':
        return False
    if not len(parts) > 0:
        return False
    if accumulation.strip() != '':
        return False
    previous_node = get_node_from_simple_text(parts[-1][1])
    if not _is_environment_node(previous_node):
        return False
    return previous_node.environmentname in display_names


# %% ../../nbs/16_latex.convert.ipynb 93
DEFAULT_ENVIRONMENTS_TO_NOT_DIVIDE_ALONG = [
    'equation', 'equation*', 'proof', 'align', 'align*', 'enumerate', 'itemize', 'label',
    'eqnarray', 'quote', 'tabular', 'table']
def divide_latex_text(
        document: str, 
        # environments_to_divide_along: list[str], # A list of the names of environments that warrant a new note
        # numbered_environments: list[str], # A list of the names of environments which are numbered in the latex code. 
        environments_to_not_divide_along: list[str] = DEFAULT_ENVIRONMENTS_TO_NOT_DIVIDE_ALONG, # A list of the names of the environemts along which to not make a new note, unless the environment starts a section (or the entire document).
        ) -> list[tuple[str, str]]: # Each tuple is of the form `(note_title, text)`, where `note_title` often encapsulates the note type (i.e. section/subsection/display text of a theorem-like environment) along with the numbering and `text` is the text of the part. Sometimes `title` is just a number, which means that `text` is not of a `\section` or `\subsection` command and not of a theorem-like environment.
    r"""Divide LaTeX text to convert into Obsidian.md notes.

    Assumes that the counters in the LaTeX document are either the
    predefined ones or specified by the `\newtheorem` command.

    This function does not divide out `\subsubsection`'s.

    Proof environments are assigned to the same parts their prcededing
    theorem-like environments, if available.

    TODO: Implement counters specified by `\newcounter`, cf. 
    https://www.overleaf.com/learn/latex/Counters#LaTeX_commands_for_working_with_counters.
    """
    numbertheorem_counters = numbered_newtheorems_counters_in_preamble(document)
    explicit_numberwithins = numberwithins_in_preamble(document)
    numberwithins = _setup_numberwithins(explicit_numberwithins, numbertheorem_counters)
    all_numberwithins = _setup_all_numberwithins(explicit_numberwithins, numbertheorem_counters)
    # environments_to_counters = counters_for_environments(document)
    display_names = display_names_of_environments(document)
    counters = _setup_counters(numbertheorem_counters)
    unnumbered_environments = _unnumbered_environments(
        numbertheorem_counters, display_names)
    # Eventually gets returned
    preamble, main_document = divide_preamble(document)
    document_node = find_document_node(main_document)
    swap_numbers = swap_numbers_invoked(preamble)
    parts = []
    # "Accumulates" a "part" until text that should comprise a new part is encountered
    accumulation = '' 
    for node in document_node.nodelist:
        accumulation = _process_node(
            node, environments_to_not_divide_along, accumulation,
            numbertheorem_counters,
            numberwithins, all_numberwithins, counters,
            display_names, swap_numbers, parts)
    _append_non_environment_accumulation_to_parts_if_non_empty(
        accumulation, counters, parts)
    return parts


def _process_node(
        node, environments_to_not_divide_along, accumulation,
        numbertheorem_counters,
        numberwithins, all_numberwithins, counters,
        display_names, swap_numbers, parts) -> str:
    """
    Update `accumulation`, `counter`, and `parts` based on the contents of `node`.

    Also return 'accumulation` to update it.

    This is a helper function for `divide_latex_text`.

    """
    # If node is a proof immediately following a theorem-like environment
    # Then add it to said theorem-like environment
    if _node_is_proof_immediately_following_a_theorem_like_environment(
            node, accumulation, parts, display_names):
        parts[-1][1] += node.latex_verbatim()
        return accumulation

    _change_counters(node, counters, numbertheorem_counters, numberwithins)
    if _node_warrants_own_part(
            node, environments_to_not_divide_along, accumulation, parts):
        accumulation =  _append_non_environment_accumulation_to_parts_if_non_empty(
            accumulation, counters, parts)
        
        title = _title(
            node, numbertheorem_counters, numberwithins, all_numberwithins,
            display_names, counters, swap_numbers).strip()
        title = title.replace('\n', '') 
        parts.append([title, node.latex_verbatim()])
    else:
        accumulation += node.latex_verbatim()
        # In _change_counters`, the '' counter is incremented by default.
        # This offsets the incorrectly incrementation.
    return accumulation


def _append_non_environment_accumulation_to_parts_if_non_empty(
        accumulation: str, counters, parts):
    """Append accumulation to `parts` if `accumulation` is nonempty
    and return the updated `accumulation` """
    if accumulation.strip() != '':
        counters[''] += 1
        parts.append([str(counters['']).strip(), accumulation.strip()])
        return ''
    else:
        return accumulation.strip()






# %% ../../nbs/16_latex.convert.ipynb 113
def _part_starts_section(
        part: tuple[str, str]):
    """
    Return `True` if `part` starts a section (explicitly),
    cf. `divide_latex_text`.
    """
    return part[1].startswith(r'\section')
    # node = get_node_from_simple_text(part[1])
    # return _is_section_node(node)


def _part_starts_subsection(
        part: tuple[str, str]):
    """Return `True` if `part` starts a subsection, cf. `divide_latex_text`."""
    return part[1].startswith(r'\subsection')
    # node = get_node_from_simple_text(part[1])
    # return _is_subsection_node(node)

# %% ../../nbs/16_latex.convert.ipynb 115
UNTITLED_SECTION_TITLE = 'Untitled Section'
def section_and_subsection_titles_from_latex_parts(
        parts: list[tuple[str, str]], # An output of `divide_latex_text`
        # verbose_sections: bool = False, # 
        # short_subsections: bool = False,
        # section_name: str = 'section',
        # subsection_name: str = 'subsection')\
        ) -> list[list[str]]: # Each list corresponds to a section. The first entry of the list is the title of the section and the other entries are the titles of the subsections. 
    """
    Return a list of lists of titles for the sections and subsections in `parts`

    Unnumbered sections get their own list. Unnumbered subsections are also included in lists.
    All the titles are striped (of leading and trailing whitespaces).
    """
    sections_and_subsections = []
    for part in parts:
       _consider_part_to_add(part, sections_and_subsections) 
    return sections_and_subsections


def _consider_part_to_add(
        part: list[tuple[str, str]],
        sections_and_subsections: list[list[str]]):
    """Add the title of `part` to `sections_and_subsections`
    if `part` starts a section or subsection."""
    title = part[0].strip()
    if _part_starts_section(part):
        sections_and_subsections.append([title])
    elif _part_starts_subsection(part):
        sections_and_subsections[-1].append(title)
    elif not sections_and_subsections:
        # If sections and subsections is empty and the very first `part`
        # does not explicitly start a section, then we are in an untitled
        # section.
        sections_and_subsections.append([UNTITLED_SECTION_TITLE])
        


# %% ../../nbs/16_latex.convert.ipynb 130
def custom_commands(
        preamble: str, # The preamble of a LaTeX document.
        ) -> list[tuple[str, int, Union[str, None], str]]: # Each tuple consists of 1. the name of the custom command 2. the number of parameters 3. The default argument if specified or `None` otherwise, and 4. the display text of the command.
    """
    Return a dict mapping commands (and math operators) defined in `preamble` to
    the number of arguments display text of the commands.

    Assumes that the newcommands only have at most one default parameter (newcommands with
    multiple default parameters are not valid in LaTeX).

    Ignores all comented newcommands.
    """
    preamble = remove_comments(preamble)
    latex_commands = _commands_from_newcommand_and_declaremathoperator(preamble)
    # tex_commands = _commands_from_def(preamble)
    return latex_commands


def _commands_from_newcommand_and_declaremathoperator(
        preamble: str, # The preamble of a LaTeX document
        ) -> list[tuple[str, int, Union[str, None], str]]: # Each tuple consists of 1. the name of the custom command 2. the number of parameters 3. The default argument if specified or `None` otherwise, and 4. the display text of the command.
    """
    Get custom commands from invocations of `\newcommand` and `DeclareMathOperator`
    in the preamble.
    """
    # newcommand_regex = regex.compile(
    #     r'(?<!%)\s*\\(?:(?:re)?newcommand|DeclareMathOperator)\s*\{\\\s*(\w+)\s*\}\s*(?:\[(\d+)\]\s*(?:\[(\w+)\])?)?\s*\{((?>[^{}]+|\{(?4)\})*)\}', re.MULTILINE)
    newcommand_regex = regex.compile(
        r'(?<!%)\s*\\(?:(?:re)?newcommand|DeclareMathOperator)\s*(?:\{\\\s*(\w+)\s*\}|\\\s*(\w+))\s*(?:\[(\d+)\]\s*(?:\[(\w+)\])?)?\s*\{((?>[^{}]+|\{(?5)\})*)\}', re.MULTILINE)

    commands = []
    for match in newcommand_regex.finditer(preamble):
        name_surrounded_in_parentheses = match.group(1) # e.g. \newcommand{\A}
        name_without_parentheses = match.group(2) # e.g. \newcommand\A
        num_args = match.group(3)
        optional_default_arg = match.group(4)
        definition = match.group(5)

        if name_surrounded_in_parentheses is not None:
            name = name_surrounded_in_parentheses
        else:
            name = name_without_parentheses

        # Convert the number of arguments to an integer, if it was specified
        if num_args is not None:
            num_args = int(num_args)
        else:
            num_args = 0

        commands.append((name, num_args, optional_default_arg, definition))
    return commands


# def _commands_from_def(
#         preamble: str
#         ) -> list[tuple[str, int, Union[str, None], str]]: # Each tuple consists of 1. the name of the custom command 2. the number of parameters 3. The default argument if specified or `None` otherwise, and 4. the display text of the command.
#     """
#     """
#     def_regex = regex.compile(
#         r'(?<!%)\s*\\def\s*(\\[a-z0-9]+)'
#     )


# %% ../../nbs/16_latex.convert.ipynb 133
def regex_pattern_detecting_command(
        command_tuple: tuple[str, int, Union[None, str], str], # Consists of 1. the name of the custom command 2. the number of parameters 3. The default argument if specified or `None` otherwise, and 4. the display text of the command.
        ) -> regex.Pattern:
    """Return a `regex.pattern` object (not a `re.pattern` object) detecting
    the command with the specified number of parameters, optional argument,
    and display text.

    Assumes that the curly braces used to write the invocations of the commands
    are balanced and properly nested. Assumes that there are no two commands
    of the same name.
    """
    command_name, num_parameters, optional_arg, _ = command_tuple
    backslash_name = fr"\\{command_name}"
    optional_argument_detection = fr"(?:\[(.*?)\])?" if optional_arg is not None else ""
    argument_detection = r""
    if optional_arg is not None:
        trailing_arguments = [_argument_detection(i) for i in range(2, 1+num_parameters)]
        trailing_args_pattern = "\\s*".join(trailing_arguments)
        pattern = (f"{backslash_name}\\s*{optional_argument_detection}\\s*{trailing_args_pattern}")
    elif num_parameters > 0:
        arguments = [_argument_detection(i) for i in range(1, 1+num_parameters)]
        args_pattern = "\\s*".join(arguments)
        pattern = f"{backslash_name}\\s*{args_pattern}"
    else:
        # Match the command name exactly without letters immediately following
        # (but underscores following are okay).
        pattern = f"{backslash_name}(?![^\W_])"
    return regex.compile(pattern)

def _argument_detection(group_num: int):
    return "\{((?>[^{}]+|\{(?1)\})*)\}".replace("1", str(group_num))
    

# %% ../../nbs/16_latex.convert.ipynb 135
def replace_command_in_text(
        text: str,
        command_tuple: tuple[str, int, Union[None, str], str], # Consists of 1. the name of the custom command 2. the number of parameters 3. The default argument if specified or `None` otherwise, and 4. the display text of the command.
    ):
    """
    Replaces all invocations of the specified command in `text` with the display text
    with the arguments used in the display text.

    Assumes that '\1', '\2', '\3', etc. are not part of the display text. 
    """
    command_name, num_parameters, optional_arg, display_text = command_tuple
    command_pattern = regex_pattern_detecting_command(command_tuple)
    replace_pattern = display_text.replace('\\', r'\\')
    # if optional_arg is not None:
    #     replace_pattern = replace_pattern.replace('#1', optional_arg)
    replace_pattern = re.sub(r'#(\d)\b', r'\\\1', replace_pattern)
    text = regex.sub(
        command_pattern,
        lambda match: _replace_command(match, command_tuple, command_pattern, replace_pattern),
        text)
    return text

def _replace_command(
        match: regex.match,
        command_tuple: tuple[str, int, Union[None, str], str],
        command_pattern: regex.Pattern,
        replace_pattern: re.Pattern) -> str:
    """Replace the matched command with the display text"""
    command_name, num_parameters, optional_arg, display_text = command_tuple
    start, end = match.span()
    matched_string_to_replace = match.string[start:end]
    if len(match.groups()) > 0 and match.group(1) is None:
        replace_pattern = replace_pattern.replace(r'\1', optional_arg)
        replaced_string = regex.sub(command_pattern, replace_pattern, matched_string_to_replace)
        return replaced_string
    else:
        return regex.sub(command_pattern, replace_pattern, matched_string_to_replace)



# %% ../../nbs/16_latex.convert.ipynb 137
def replace_commands_in_text(
        text: str, # The text in which to replace the commands. This should not include the preamble of a latex document.
        command_tuples: tuple[str, int, Union[None, str], str], # An output of `custom_commands`. Each tuple Consists of 1. the name of the custom command 2. the number of parameters 3. The default argument if specified or `None` otherwise, and 4. the display text of the command.
    ) -> str:
    """
    Replaces all invocations of the specified commands in `text` with the
    display text with the arguments used in the display text.

    Assumes that '\1', '\2', '\3', etc. are not part of the display text. 
    """
    for command_tuple in command_tuples:
        text = replace_command_in_text(text, command_tuple)
    return text

# %% ../../nbs/16_latex.convert.ipynb 139
def replace_commands_in_latex_document(
        docment: str
        ) -> str:
    """Return the latex document (without the preamble) with invocations
    of custom commands/operators replaced with their display text.

    Assumes that all custom commands and operators are defined in the
    preamble.

    Assumes that, if commands with the same name are defined multiple times,
    only the finally defined command is used. 

    Even replaces these invocations incommented out text.
    """
    preamble, document = divide_preamble(docment)
    commands = custom_commands(preamble)
    # Note that `command_tuple[0]` is the name of the command.
    unique_commands = {command_tuple[0]: command_tuple for command_tuple in commands} 
    for _, command_tuple in unique_commands.items():
        document = replace_command_in_text(document, command_tuple)
    return document
    

# %% ../../nbs/16_latex.convert.ipynb 143
# TODO: give the option to replace quotations ``'' and `enquote`, e.g. ```unlikely intersections''` into `"unlikely intersections"`
# TODO: give the option to replace emph with `****`, e.g. ``\emph{special}``.
def adjust_common_syntax_to_markdown(
        text) -> str:
    """
    Adjust some common syntax, such as math mode delimiters and equation/align
    environments, for Markdown.

    Assumes that the tokens for math mode delimiters (e.g. `\( \)` and `\[ \]`)
    are not used otherwise.
    """
    # TODO: see if I need to add more substitutions.
    text = re.sub(r'\\\(|\\\)', '$', text)
    text = re.sub(r'\\\[|\\]', '$$', text)
    text = re.sub(r'(\\begin\{(?:align|equation|eqnarray)\*?\})', r'$$\1', text)
    text = re.sub(r'(\\end\{(?:align|equation|eqnarray)\*?\})', r'\1$$', text)
    return text

# %% ../../nbs/16_latex.convert.ipynb 146
def _replace_custom_commands_in_parts(
        parts: list[tuple[str, str]],
        custom_commands: list[tuple[str, int, Union[str, None], str]]
        ) -> list[tuple[str, str]]:
    return [(title, replace_commands_in_text(text, custom_commands))
            for title, text in parts]


def _adjust_common_syntax_to_markdown_in_parts(
        parts: list[tuple[str, str]]
        ) -> list[tuple[str, str]]:
    return [(title, adjust_common_syntax_to_markdown(text))
            for title, text in parts]

# %% ../../nbs/16_latex.convert.ipynb 148
def _adjust_common_section_titles_in_parts(
        parts: list[tuple[str, str]],
        reference_name: str):
    """Adjust common section titles in `parts`

    Common section titles include, but are not limited to,
    `Introduction`, `Notations`,
    `Conventions`, `Preliminaries`, and `Notations and Conventions`    
    
    This is a helper function for `setup_reference_from_latex_parts`.

    """
    return [(_adjusted_title(title, text, reference_name), text)
            for title, text in parts]


# TODO: also adjust title if the title is of the form
# <section_number>_<common_section_title>, e.g.
# 7_acknowledgements
# 8_references
COMMON_SECTION_TITLES = [
    'introduction', 'notations', 'conventions', 'preliminaries',
    'notations and conventions', 'definitions', 'definitions and notations',
    'references', 'acknowledgements']


def _adjusted_title(
        title: str,
        text: str,
        reference_name: str):
    """Adjust the given title"""
    node = get_node_from_simple_text(text)
    if not _is_section_node(node):
        return title
    _, section_title = _section_title(text)
    if section_title.lower() in COMMON_SECTION_TITLES:
        return f'{title}_{reference_name}'
    else:
        return title 

# %% ../../nbs/16_latex.convert.ipynb 151
def _create_notes_from_parts(
        parts: list[tuple[str, str]],
        chapters: list[list[str]],
        index_note: VaultNote, # The index note of the reference that was created.
        vault: Path,
        reference_folder: Path,
        reference_name: str,
        template_mf: MarkdownFile, # The template of the reference that was created.
        ):
    """Create notes for the vault from `parts`."""
    # headings_folder_correspondence = correspond_headings_with_folder(
    #     index_note, vault)
    title_numbering_folder_map = {
        title: convert_title_to_folder_name(title)
        for section in chapters
        for title in section}

    current_section, current_subsection = chapters[0][0], '' # section/subsection titles
    # Dict of dict of list of str. The top level keys
    # are section titles and the corresponding values are dicts whose
    # keys are subsection titles and values are lists of bulleted links texts
    # of the form `- [[linke_to_note]], Title/identifying information` to add.
    links_to_make = {current_section: {current_subsection: []}}  
    for part in parts:
        current_section, current_subsection = _create_part_or_update(
            part, title_numbering_folder_map, vault, reference_folder,
            reference_name, template_mf, current_section, current_subsection,
            links_to_make)

    _make_links_in_index_notes(
        links_to_make, title_numbering_folder_map, vault,
        reference_folder, reference_name)
    

def _make_links_in_index_notes(
        links_to_make: dict[str, dict[str, list[str]]],
        title_numbering_folder_map: dict[str, tuple[str, str]],
        vault: Path,
        reference_folder: Path,
        reference_name: str,
        ):
    for section_title, section_dict in links_to_make.items():
        section_folder = title_numbering_folder_map[section_title]
        _make_links_in_index_note_for_section(
            section_title, section_dict, section_folder,
            vault, reference_folder, reference_name)


def _make_links_in_index_note_for_section(
        section_title: str,
        section_dict: dict[str, list[str]],
        section_folder: str,
        vault: Path,
        reference_folder: Path,
        reference_name: str):
    rel_path = reference_folder / section_folder / f'_index_{section_folder}.md'
    section_index_note = VaultNote(vault, rel_path=rel_path)
    mf = MarkdownFile.from_vault_note(section_index_note)
    for subsection_title, links_to_make_in_header in section_dict.items():
        mf.add_line_in_section(
            subsection_title,
            {'type': MarkdownLineEnum.UNORDERED_LIST,
             'line': '\n'.join(links_to_make_in_header)})
    mf.write(section_index_note)


def _create_part_or_update(
        part: tuple[str, str],
        title_numbering_folder_map: dict[str, tuple[str, str]],
        vault: Path,
        reference_folder: Path,
        reference_name: str,
        template_mf: MarkdownFile,
        current_section: str,
        current_subsection: str,
        links_to_make: dict[str, dict[str, list[str]]],
        ) -> tuple[str, str]:
    """
    Consider `part` for creating a new note or for updating
    `current_section` and `current_subsection`

    Also append to `links_to_make` for each note that is created.
    """
    if _part_starts_section(part):
        current_section = part[0]
        current_subsection = ''
        links_to_make[current_section] = {'': []}
        folder = title_numbering_folder_map[current_section]
        return current_section, current_subsection
    elif _part_starts_subsection(part):
        current_subsection = part[0]
        links_to_make[current_section][current_subsection] = []
        return current_section, current_subsection

    created_note = _create_note_for_part(
        part, title_numbering_folder_map, vault, reference_folder,
        reference_name, template_mf, current_section, current_subsection)

    _update_links_to_make(
        part, current_section, current_subsection, links_to_make,
        created_note)
    return current_section, current_subsection


def _create_note_for_part(
        part: tuple[str, str],
        title_numbering_folder_map: dict[str, tuple[str, str]],
        vault: Path,
        reference_folder: Path,
        reference_name: str,
        template_mf: MarkdownFile,
        current_section: str,
        current_subsection: str
        ) -> VaultNote: # The created VaultNote.
    """Create a note for the part"""
    note_title = sanitize_filename(part[0])
    note_contents = part[1]
    mf = template_mf.copy(True)
    mf.add_line_in_section(
        'Topic[^1]', {'type': MarkdownLineEnum.DEFAULT, 'line': note_contents})
    mf.parts[-1]['line'] += note_title
    section_folder = title_numbering_folder_map[current_section]
    # TODO: Make it so that unique_note_name indicates an unnumbered
    # note as unnumbered.
    unique_note_name = VaultNote.unique_name(
        f"{reference_name}_{note_title}", vault)
    if current_subsection == '':
        rel_path = (
            reference_folder / section_folder / f"{unique_note_name}.md")
    else:
        subsection_folder = title_numbering_folder_map[current_subsection]
        rel_path = (
            reference_folder / section_folder / subsection_folder / f"{unique_note_name}.md")
    vn = VaultNote(vault, rel_path=rel_path)
    vn.create()
    mf.write(vn)
    return vn



def _update_links_to_make(
        part: tuple[str, str],
        current_section: str,
        current_subsection: str,
        links_to_make: dict[str, dict[str, list[str]]],
        created_note: VaultNote
        ) -> None:
    """Update `links_to_make` after note is created."""
    # if current_subsection is not None:
    #     current_subsection_key = current_subsection
    # else:
    #     current_subsection_key = ''
    note_title = part[0]
    links_to_make[current_section][current_subsection].append(
        f'- [[{created_note.name}]], {note_title}'
    )
    




# %% ../../nbs/16_latex.convert.ipynb 152
# TODO: test parts without a subsection.
# TODO: somehow contents before a section are not inclued. Fix this bug.
# TODO: If section titles are completely empty, e.g. https://arxiv.org/abs/math/0212208,
# Make section titles based on reference name.
# TODO: give the option to not included commented out content from latex files.
def setup_reference_from_latex_parts(
        parts: list[tuple[str, str]], # Output of `divide_latex_text`
        custom_commands: list[tuple[str, int, Union[str, None], str]], # Output of `custom_commands` applied to the preamble of the LaTeX ddocument.`
        vault: PathLike, # An Obsidian.md vault,
        location: PathLike, # The path to make the new reference folder. Relative to `vault`.
        reference_name: PathLike, # The name of the new reference.
        authors: Union[str, list[str]], # Each str is the family name of each author.
        author_folder: PathLike = '_mathematicians', # The directory where the author files are stored in. Relative to `vault`.
        references_folder: PathLike = '_references', # The directory where the references files are stored in. Relative to `vault`.
        templates_folder: PathLike = '_templates', # The directory where the template files are stored in. Relative to `vault`.
        template_file_name: str = '_template_common', # The template file from which to base the template file of the new reference.
        notation_index_template_file_name: str = '_template_notation_index', # The template file from which to base the notation index file of the new reference.
        glossary_template_file_name: str = '_template_glossary', # The template file from which to base the glossary file of the new reference.
        setup_temp_folder: bool = True, # If `True`, creates a `_temp` folder with an index file. This folder serves to house notes auto-created from LaTeX text files before moving them to their correct directories. Defaults to `True`.
        make_second_template_file_in_reference_directory: bool = True, # If `True`, creates a copy of the template note within the directory for the reference.
        copy_obsidian_configs: Optional[PathLike] = '.obsidian', # The folder relative to `vault` from which to copy obsidian configs.  If `None`, then no obsidian configs are copied to the reference folder. Defaults to `.obsidian`. 
        overwrite: Union[str, None] = None, # Specifies if and how to overwrite the reference folder if it already exists.  - If `'w'`, then deletes the contents of the existing reference folder, as well as the template and reference file before setting up the reference folder before creating the new reference folder.  - If `'a'`, then overwrites the contents of the reference folder, but does not remove existing files/folders.  - If `None`, then does not modify the existing reference folder and raises a `FileExistsError`.
        confirm_overwrite: bool = True, # Specifies whether or not to confirm the deletion of the reference folder if it already exists and if `overwrite` is `'w'`. Defaults to `True`.
        verbose: bool = False,
        replace_custom_commands: bool = True, # If `True`, replace the custom commands in the text of `parts` when making the notes.
        adjust_common_latex_syntax_to_markdown: bool = True, # If `True`, apply `adjust_common_syntax_to_markdown` to the text in `parts` when making the notes.`
        ) -> None:
    """Set up a reference folder in `vault` using an output of `divide_latex_text`, create
    notes from `parts`, and link notes in index files in the reference folder.

    Assumes that

    - `parts` is derived from a LaTeX document in which
        - all of the text belongs to sections.
        - all of the sections/subsections are uniquely named
    - The template file is has a section `# Topic`
    - The last line of the template file is a footnote indicating where the note comes from.
    - There is at most one reference folder in the vault whose name is given by
      `reference_name`.

    `parts` itself is not modified, even if `replace_custom_commands` and/or
    `adjust_common_latex_syntax_to_markdown` are set to `True`.

    cf. `setup_folder_for_new_reference` for how the reference folder is set up..

    The names for the subfolders of the reference folder are the section titles, except
    for sections with common titles such as `Introduction`, `Notations`, `Conventions`,
    `Preliminaries`, and `Notations and Conventions`. This ensures that the index
    file names for sections in different reference folders do not have the same name.

    Text/parts that precede explicitly given sections are included in the 
    first section's folder and are linked in the first section's index file.
    """
    parts = _adjust_common_section_titles_in_parts(parts, reference_name)
    chapters = section_and_subsection_titles_from_latex_parts(parts)
    if chapters[0][0] == UNTITLED_SECTION_TITLE:
        chapters[0][0] = f'{reference_name} {UNTITLED_SECTION_TITLE}'
    setup_folder_for_new_reference(
        reference_name, location, authors, vault, author_folder,
        references_folder, templates_folder, template_file_name,
        notation_index_template_file_name, 
        glossary_template_file_name, chapters, setup_temp_folder,
        make_second_template_file_in_reference_directory,
        copy_obsidian_configs, overwrite, confirm_overwrite, verbose)
    index_note = VaultNote(
        vault, rel_path=Path(location) / reference_name / f'_index_{reference_name}.md')
    template_note = VaultNote(vault, name=f'_template_{reference_name}')
    template_mf = MarkdownFile.from_vault_note(template_note)

    if replace_custom_commands:
        parts = _replace_custom_commands_in_parts(parts, custom_commands)
    if adjust_common_latex_syntax_to_markdown:
        parts = [(title, adjust_common_syntax_to_markdown(text))
                 for title, text in parts]
    
    reference_folder = Path(location) / reference_name
    _create_notes_from_parts(
        parts,
        chapters, 
        index_note,
        vault,
        reference_folder,
        reference_name,
        template_mf)
    




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/17_markdown.obsidian.personal.machine_learning.__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../../../../../nbs/17_markdown.obsidian.personal.machine_learning.__init__.ipynb 2
from bisect import bisect_left, bisect_right
from os import PathLike
from random import sample, choices
import re
from typing import Union

# from fastai.text.all import *
import pandas as pd

from .....helper import latex_indices, find_regex_in_text, current_time_formatted_to_minutes
from ....markdown.file import MarkdownFile
from trouver.markdown.obsidian.footnotes import (
    remove_footnote_mentions_in_markdown_text
)
# TODO: import the rest



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/03_markdown.obsidian.vault.ipynb.

# %% auto 0
__all__ = ['NoteNotUniqueError', 'NoteDoesNotExistError', 'NotePathIsNotIdentifiedError', 'path_to_obs_id',
           'all_paths_to_notes_in_vault', 'all_note_paths_by_name', 'note_path_by_name', 'note_name_unique',
           'note_name_from_path', 'VaultNote']

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 3
from pathlib import Path
import os
from os import PathLike
from trouver.helper import (
    path_name_no_ext, path_no_ext
)
from typing import Union

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 7
class NoteNotUniqueError(FileNotFoundError):
    """
    A `NoteNotUniqueError` is raised when a `VaultNote` is specified
    by name and not by relative path in the vault, but
    the vault is found to have multiple notes of the name.

    **Attributes**

    - `note_name` - str
        - The name of the note which is not unique in the vault.
    - `notes` - list[str]
        - The paths of the notes whose names are `note_name`.
    """
    def __init__(self, /, *args, **kwargs):
        super().__init__(*args, **kwargs)
            
    @classmethod
    def from_note_names(cls, note_name: str, notes: list[str]):
        """Construct a `NoteNotUniqueError` object from note names"""
        return cls(
            f'The note of the following name is not unique: {note_name}\n'\
            f'The name points to the following files: {notes}')



# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 9
class NoteDoesNotExistError(FileNotFoundError):
    """
    A `NoteDoesNotExistError` is raised when a `VaultNote` is specified
    by either name and or by relative path in the vault, but
    the vault is found to have no notes of the name.
    """
    def __init__(self, /, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_note_name(cls, note_name: str):
        """Construct a `NoteDoesNotExistError` object from note name"""
        return cls(
            f'The note of the following name does not exist: {note_name}.'
            f' Make sure that the argument passed to `note_name` does not'
            f' erroneously end with `.md`, e.g. pass `this_is_a_note`'
            f' instead of `this_is_a_note.md`.')

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 11
class NotePathIsNotIdentifiedError(RuntimeError):
    """
    A `NotePathIsNotIdentifiedError` is raised when the `rel_path` attribute of a
    `VaultNote` object is expected to be identified (i.e. a path and not `None`) but
    this expectation is not fulfilled.
    """
    def __init__(self, /, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_note(
            cls,
            note):
        """Construct a `NotePatahIsNotIdentifiedErrro` object from a `VaultNote`."""
        return cls(f'The `rel_path` attribute of the `VaultNote` object is expected'
                   f' to be identified, and yet it is not. The vault of the'
                   f' `VaultNote` object is {note.vault} and the name of the object'
                   f' is {note.name}.')

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 14
def path_to_obs_id(
        rel_path: PathLike # A path representation the path of an Obsidian note relative to its vault. This does not have to be an existing path.
        ) -> str: # The obsidian url of the hypothetical note within its vault. Note that this does not end with the file extension `.md`.
    """Convert a relative path of an Obsidian note to the Obsidian identifying
    str.
    
    This identification is for a vault-internal Wikilink.

    Note that this function does not have a vault as a parameter.
    """
    path_without_extension = path_no_ext(rel_path)
    return path_without_extension.replace('\\', '/')

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 21
def all_paths_to_notes_in_vault(
        vault: PathLike, as_dict: bool = False)\
        -> Union[list[str], dict[str, list[str]]]:
    """Return the paths, relative to the Obsidian vault, of notes 
    in the Obsidian vault.
       
    This may not actually return all of the paths to the notes, see
    the parameter `as_dict`.

    **Parameters**

    - `vault` - `PathLike`
        - The path to the Obsidian vault directory
    - `as_dict` - `bool`
        - If `True`, then returns a dict. If `False`, then returns
        a list. Defaults to `False`. If there are multiple notes with the same
        name in the vault, and `as_dict` is set to `True`, then the dictionary
        will contain only one of the (relative) paths to those notes among its
        values. If `as_dict` is set to `False`, then the list will contain
        all the paths to the notes even when notes with non-unique names exist.
        
    **Returns**

    - Union[list[str], dict[str, str]]
        - Each str represents the path relative to `vault`. If `as_dict` is
        True, then returns a dict whose keys are str, which are (unique) names
        of the notes in the vault, and the values are the paths.
    """
    vault = Path(vault)
    paths =  [os.path.relpath(path, vault) for path in vault.glob(f'**/*.md')]
    if as_dict:
        dicty = {path_name_no_ext(path): [] for path in paths}
        for path in paths:
            dicty[path_name_no_ext(path)].append(path)
        return dicty
    else:
        return paths


# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 27
def all_note_paths_by_name(
        name: str,  # Name of the note(s) to find
        vault: PathLike,  # The path to the Obsidian vault directory
        subdirectory: Union[PathLike, None] = None # The path to a subdirectory in the Obsidian vault, relative to `vault`. If `None`, then denotes the root of the vault.
        ) -> list[Path]: # Each item is a path to a note of the given name, relative to `vault`.
    """Return the relative paths to all notes in the Obsidian vault 
    with the specified name in the specified subdirectory.
    
    This function does not assume that the specified subdirectory in the vault
    has at most one note of the specified name.
    
    """
    vault = vault if vault != None else ''
    vault = Path(vault)
    subdirectory = subdirectory if subdirectory != None else ''
    directory_path = vault / subdirectory
    all_notes_of_name = directory_path.glob(f'**/{name}.md')
    all_notes_of_name = list(all_notes_of_name)
    return [note_path.relative_to(vault) for note_path in all_notes_of_name]

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 32
# TODO: include examples of `hints` parameter.
def note_path_by_name(
        name: str, # The path to the Obsidian vault directory.
        vault: PathLike, # The path to a subdirectory in the Obsidian vault. If `None`, then denotes the root of the vault.
        subdirectory: Union[PathLike, None] = None, # The path to a subdirectory in the Obsidian vault. If `None`, then denotes the root of the vault.
        hints: Union[list[PathLike], None] = None # Hints of which directories, relative to `subdirectory` that the note may likely be in. This is for speedup. The directories will be searched in the order listed.
        ) -> Path: # The note of the specified name in the specified subdirectory of the vault.
    """Return the path, relative to a subdirectory in the vault, 
    of the note of the specified name.

    **Raises**

    - NoteNotUniqueError
        - If the note of the specified name is not unique in the subdirectory.
    - NoteDoesNotExistError
        - If the note of the specified name does not exist in the subdirectory.

    **See Also**

    - The constructor of the `VaultNote` class
        - passing an argument to the `name` parameter of this constructor
        method essentially does the same thing as this function, except the
        constructor method uses a cache.
    """
    if not subdirectory:
        subdirectory = ''
    if not hints:
        hints = []
    vault = Path(vault)
    subdirectory = Path(subdirectory)
    # absolute_subdirectory = vault / subdirectory
    hints.append('')  # Search in subdirectory if all else fails
    for hint in hints:
        search_results = all_note_paths_by_name(name, vault, subdirectory / hint)
        if len(search_results) > 1:
            raise NoteNotUniqueError.from_note_names(name, search_results)
        elif len(search_results) == 1:
            return search_results[0]
    raise NoteDoesNotExistError.from_note_name(name)

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 41
def note_name_unique(
        name: str, # Name of the note.
        vault: PathLike # Path to the vault.
        ) -> bool: 
    """Return `True` if a note of the specified name exists and 
    is unique in the Obsidian vault.
    """
    return len(all_note_paths_by_name(name, vault)) == 1

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 45
def note_name_from_path(
        note_path: str # The path of the note. The note does not need to exist.
        ) -> str: # The name of the note.
    """Return the name of a note from its path.
    """
    return path_name_no_ext(note_path)

# %% ../../../nbs/03_markdown.obsidian.vault.ipynb 48
# TODO: change NoteDoesNotExistError to NoteNotFoundError if not found in Cache.
# TODO: when making NoteNotFoundError, add it to
# markdown.obsidian.personal.reference.delete_reference_folder

# TODO: modify the constructor so that it does not need the note of the specified name to exist
# TODO: look at parts where the cache is accessed/modified. Factor them out into getter/setter methods
# taking in the vault and name as inputs.
# TODO: look at how NoteDoesNotExistError is used and make modifications to account for the modification
# of the constructor.
# TODO: test hidden methods
class VaultNote:
    """Represents a note in an Obsidian vault, without regards to the contents.
    
    The note does not have to exist, except in circumstances stating 
    otherwise.

    TODO go through the methods of this class to see which methods assume that
    the note exists and which do not.

    TODO finish the sentence below.
    A `VaultNote` can be specified by either the `rel_path` or the `name` argument
    in its constructor. If `name` is specified, then the 
    
    TODO implement subdirectory hint
    
    **Attributes**

    - vault - Path
        - The (relative or absolute) path of the Obsidian vault
        that the note is located in.
    - name - str
        - The name of the note in the vault.
    - rel_path - str
        - The note's path relative to the vault. If 
    - cache - dict[str, dict[str, list[str]]], class attribute
        - The keys are string, which are paths to vaults. The
        corresponding values are dict whose keys are string, which are
        names in the vault of the (unique) note of that name, and whose
        values are list of string, which are paths to the note relative to the
        vault. The cache is not automatically updated when notes are
        moved.
    
    **Parameters**

    - vault - PathLike
    - rel_path - PathLike
    - name - str
        - The name of the note in the vault. Defaults to the empty str.
            - If `None`, then the `rel_path` parameter should be used 
            to determine `self.name` instead. 
            - If not `None`, then the note must uniquely exist in the
            vault.
    - `subdirectory` - Union[PathLike, None]
    - `hints` - list[PathLike]

    **Raises**
    
    - ValueError
        - if `rel_path` and `name` are both `None`.
    """
    
    cache = {}
    
    # TODO: make the constructor take an optional update_cache parameter.
    def __init__(
            self,
            vault: PathLike, # The (relative or absolute) path of the Obsidian vault that the note is located in.
            rel_path: PathLike = None, # The note's path relative to the vault. If `None`, then the `name` parameter is used to determine the note instead. Defaults to `None`.
            name: str = None, # The name of the note. If `None`, then the `rel_path` parameter is used to determine the note instead. Defaults to `None` 
            subdirectory: Union[PathLike, None] = '', # The relative path to a subdirectory in the Obsidian vault. If `None`, then denotes the root of the vault. Defaults to the empty str. 
            hints: list[PathLike] = [] # Paths, relative to `subdirectory`, to directories where the note file may be found. This is for speedup. Defaults to the empty list, in which case the vault note is searched in all of `subdirectory`.
            ):
        self.vault = Path(vault)
        if rel_path is None and name is None:
            raise ValueError(
                "In constructing a `VaultNote` object, the parameters `rel_path`"
                " and `name` parameters were expected to be given arguments, but"
                " both parameters are given `None` as arguments.")
        if rel_path is not None:
            self.rel_path = str(rel_path)
            self.name = note_name_from_path(self.rel_path)
        else:
            self.name = name
            self.rel_path = None
            self.identify_rel_path(update_cache=True)

    def obsidian_identifier(self) -> str:
        """Return the Obsidian identifier of the `VaultNote` object.
        
        This is the note's unqiue Obsidian id in the vault. This is like a
        path str with forward slashes `/` (as opposed to backwards `\` slashes)
        and without a file extension (`.md`).
        """
        return path_to_obs_id(self.rel_path)

    def rel_path_identified(self) -> bool:
        """Return `True` if `self.rel_path` is identified, i.e. is a path
        that is not `None`."""
        return self.rel_path is not None

    def _identify_rel_path(self) -> Union[Path, None]:
        """Returns the Path to the note that this object represents.

        More precisely, if `rel_path` is specified at construction or
        if `self.rel_path` is already identified, then this method
        returns that path. Otherwise, this method looks into the cache to
        check if a note of the `name` specified at construction is in the
        vault and returns the first note in the list of the `name` in the
        cache. If no such note exists, then this method reutrns `None`.
        """
        if self.rel_path is not None:
            return self.rel_path
        cache_search = self.__class__._get_from_cache(self.vault, self.name)
        if cache_search:  # `cache_search` could be `None` or an empty list.
            return cache_search[0]
        return None

    def identify_rel_path(
            self,
            update_cache=False # If `True`, if the cache is searched, and if a note of the specified name is not found in the cache, then the cache is updated and searched again. Defaults to `False`.
            ) -> None:
        """Sets `self.rel_path` to a path, if not already done so.

        If `self.rel_path` is not already set as a path, then the cache
        is searched to find a note whose name is `self.name` (which is
        necessarily specified).
        """
        rel_path = self._identify_rel_path()
        if rel_path is not None:
            self.rel_path = rel_path
        elif update_cache:
            self.__class__.update_cache(self.vault)
            self.identify_rel_path(update_cache=False)

    @classmethod
    def _get_from_cache(
            cls,
            vault: PathLike,
            name: str) -> Union[list[str], None]:
        """Return the cache's list of notes of the specified name in the
        specified vault.

        If no such list exists in the cache, then return `None`.
        """
        vault = str(vault)
        if vault not in cls.cache:
            return None
        vault_dict = cls.cache[vault]
        if not name in vault_dict:
            return None
        return vault_dict[name]
                        
    def exists(
            self,
            update_cache=False # If `True`, then update the cache and try to identify `self.rel_path` before verifying whether the note exists in the vault.
            ) -> bool:
        """Returns `True` if `self.rel_path` is identified and
        if the note exists in the vault."""
        if self.rel_path is None:
            if update_cache:
                self.identify_rel_path(update_cache=True)
            else:
                return False
        return os.path.exists(self.path())
            
    def path(self,
             relative=False # If `True`, then return the path relative to the vault.
             ) -> Union[Path, None]: # Path to the note if self.rel_path is deterined. `None` otherwise.
        """Returns the path to the note.

        Assumes that `self.rel_path` has been identified.

        **Raises**
        - NotePathIsNotIdentifiedError
            - If the relative path of `self` is not identified.
        """
        if not self.rel_path_identified():
            raise NotePathIsNotIdentifiedError.from_note(self)
        return Path(self.rel_path) if relative\
            else self.vault / self.rel_path
    
    def directory(self,
                  relative=False # If `True`, then return the path of the directory relative to the vault.
                  ) -> Path: # The path of the directory that the note is in.
        """Return the directory that the note is in.
        """
        rel_dir = Path(os.path.dirname(self.rel_path))
        return rel_dir if relative else self.vault / rel_dir
    
    def create(self):
        """Create the note if it does not exist.
        
        The directory of the file needs to be created separately
        beforehand.

        If the file exists, then a FileExistsError is raised and
        the file modification time is not changed.
        
        **Raises**

        - FileExistsError
            - If the file already exists.
        - FileNotFoundError
            - If the directory of the file does not already exist.
        """
        Path(self.path()).touch(exist_ok=False)
        self.__class__._add_single_entry_to_cache(
            self.vault, self.rel_path)

    def delete(self):
        """Delete the note if it exists.
        
        This updates the cache if necessary
        """
        if self.exists(update_cache=True):
            os.remove(self.path())
            self.__class__._remove_single_entry_from_cache(self.vault, self.rel_path)
        
    def move_to(self,
                rel_path: PathLike # The path in which to rename the path to `self` as, relative to `self.vault`.
                ) -> None:
        """Move/rename the note to the specified location in the vault,
        assuming that it exists.
        """
        if not self.exists():
            return
        os.rename(self.path(), Path(self.vault) / rel_path)
        self.__class__._remove_single_entry_from_cache(
            self.vault, Path(self.rel_path))
        self.__class__._add_single_entry_to_cache(
            self.vault, Path(rel_path))
        self.rel_path = str(rel_path)
        self.name = note_name_from_path(self.rel_path)
        
    def move_to_folder(self,
                       rel_dir: PathLike # The path of the directory in which to move `self` to, relative to `self.vault`.
                       ) -> None:
        """Move the note to the specified folder in the vault, assuming that
        if exists.
        """
        self.move_to(Path(rel_dir) / f'{self.name}.md')
        
    def text(self
            ) -> str: # Text contained in the note, assuming that the note exists.
        """Returns the text contained in the note.
        
        **Raises**

        - NoteDoesNotExistError
            - If `self` does not point to an existing note.
        """
        if not self.exists():
            raise NoteDoesNotExistError.from_note_name(self.name)
        with open(self.vault / self.rel_path, 
                  'r', encoding='utf-8') as reader:
            text = reader.read()
            reader.close()
        return text

    @classmethod
    def update_cache(cls,
                     vault: PathLike # The vault.
                     ) -> None:
        """Class method to update cache for `vault` by inspecting all files
        in subdirectories of `vault`."""
        VaultNote.cache[str(vault)] = all_paths_to_notes_in_vault(
            vault, as_dict=True)
        
    @classmethod
    def clear_cache(cls):
        """Class method to clear out the entire cache for all vaults."""
        VaultNote.cache = {}
        
    @classmethod
    def _check_name_exists_and_unique_in_vault_cache(
            cls,
            vault: PathLike, # The vault
            name: str # The name
            ) -> None:
        """
        Raise `NoteNotUniqueError` or `NoteDoesNotExistError` if the note of
        the specified name is not unique or does not exist in the cache.
        
        **Parameters**
        - vault - PathLike
        - name - str

        **Raises**
        - `NoteNotUniqueError`
        - `NoteDoesNotExistError`.
        """
        vault_dict = cls.cache[str(vault)]
        if not name in vault_dict or len(vault_dict[name]) == 0:
            raise NoteDoesNotExistError.from_note_name(name)
        elif len(vault_dict[name]) > 1:
            raise NoteNotUniqueError.from_note_names(
                name, vault_dict[name])

    @classmethod
    def _check_if_cache_needs_to_update(
            cls, vault: PathLike, name: str) -> bool:
        """Returns `True` if the cache needs to update by virtue of not finding
        notes of the specified `name` in the `vault`.
        """
        return not bool(cls._get_from_cache(vault, name))

    @classmethod
    def _add_single_entry_to_cache(
            cls, vault: PathLike, rel_path: PathLike) -> None:
        """Adds a single entry for a note to the cache.
        
        Does nothing if the entry already exists.

        **Parameters**
        - vault - PathLike
        - rel_path - PathLike
            - The path to the note, relative to `vault`.
        """
        vault_str = str(vault)
        if vault_str not in cls.cache:
            cls.cache[vault_str] = {}
        name = note_name_from_path(rel_path)
        if name not in cls.cache[vault_str]:
            cls.cache[vault_str][name] = []
        if rel_path not in cls.cache[vault_str][name]:
            cls.cache[vault_str][name].append(str(rel_path))
    
    @classmethod
    def _remove_single_entry_from_cache(
            cls, vault: PathLike, rel_path: PathLike) -> None:
        """Removes a single entry for a note from the cache.

        Does nothing if the entry is not already there.

        **Parameters**
        - vault - PathLike
        - rel_path - PathLike
            - The path to the note, relative to `vault`.
        """
        vault_str = str(vault)
        if vault_str not in cls.cache:
            return
        name = note_name_from_path(rel_path)
        if name not in cls.cache[vault_str]:
            return
        rel_path = Path(rel_path)
        cls.cache[vault_str][name] = [
            cache_path for cache_path in cls.cache[vault_str][name]
            if Path(cache_path) != rel_path]

    @classmethod
    def unique_name(
            cls,
            name: str, # The base name for the note.
            vault: PathLike # The vault
            ) -> str: # A str obtained by appending `_{some number}` to the end of `name`.
        """A class method to return a name for a note that is unique in
        the vault based on a specified name.
        """
        if not str(vault) in VaultNote.cache:
            cls.update_cache(vault)
        if not name in VaultNote.cache[str(vault)]:
            return name
        i = 1
        while f'{name}_{i}' in VaultNote.cache[str(vault)]:
            i += 1
        return f'{name}_{i}'



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/18_markdown.obsidian.personal.machine_learning.notations.ipynb.

# %% auto 0
__all__ = []

# %% ../../../../../nbs/18_markdown.obsidian.personal.machine_learning.notations.ipynb 2
from deprecated import deprecated
from os import PathLike
import re
from typing import Optional

from fastai.text.all import *
import pandas as pd
import spacy
from trouver.helper import (
    current_time_formatted_to_minutes, find_regex_in_text,
    defs_and_notats_separations, replace_string_by_indices,
    notation_asterisk_indices,
    double_asterisk_indices, latex_indices
)
# from trouver.machine_learning.text.encoder_decoder import (
#     EncoderRNN, AttnDecoderRNN, train, trainIters, evaluate,
#     evaluateRandomly
# )
# from trouver.machine_learning.text.tokenize import (
#     replace_bold, replace_math_mode_strings, special_cases
# )
from ....markdown.file import MarkdownFile, MarkdownLineEnum
from trouver.markdown.obsidian.personal.note_processing import (
    process_standard_information_note
)
from ..notes import notes_linked_in_note, notes_linked_in_notes_linked_in_note
# from trouver.markdown.obsidian.personal.notation import (
#     make_a_notation_note, latex_to_path_accepted_string, make_notation_notes_from_double_asts)

from ...vault import VaultNote



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/05_markdown.markdown.heading.ipynb.

# %% auto 0
__all__ = ['heading_level', 'heading_title']

# %% ../../../nbs/05_markdown.markdown.heading.ipynb 3
def heading_level(
        heading_str: str # A str representing a markdown heading. Starts with 1 to 6 sharps `'#'`.
        ) -> int: # Between 1 and 6, inclusive.
    """Return the level of the heading string."""
    without_sharps = heading_str.lstrip('#')
    return len(heading_str) - len(without_sharps)

# %% ../../../nbs/05_markdown.markdown.heading.ipynb 5
def heading_title(
        heading_str: str # A str representing a markdown heading. Starts with 1 to 6 sharps `'#'`.
        ) -> str:
    """Return the heading title, without the starting sharps.
    """
    return heading_str.lstrip('#').strip()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/07_markdown.obsidian.tags.ipynb.

# %% auto 0
__all__ = ['to_tag_str', 'tag_without_hashtag', 'tag_is_auto_tag', 'strip_auto_from_tag']

# %% ../../../nbs/07_markdown.obsidian.tags.ipynb 3
def to_tag_str(tag_name):
    return f'#{tag_name}'

# %% ../../../nbs/07_markdown.obsidian.tags.ipynb 4
def tag_without_hashtag(tag):
    return tag[1:] if tag.startswith('#') else tag

# %% ../../../nbs/07_markdown.obsidian.tags.ipynb 5
def tag_is_auto_tag(tag):
    tag = tag_without_hashtag(tag)
    return tag.startswith('_auto/')

# %% ../../../nbs/07_markdown.obsidian.tags.ipynb 6
def strip_auto_from_tag(tag, with_hash_tag: bool = True) -> str:
    """Returns the tag without the `'_auto/'` prefix.
    
    This will essentially return the same tag if the tag does not
    have the `'_auto/'` prefix to begin with.
    
    **Parameters**
    - tag - str
    - with_has_tag - bool
        - If `True`, returns a str with a hash tag `'#'` in front.
        Defaults to `True`.
    """
    tag = tag_without_hashtag(tag)
    tag = tag[6:] if tag.startswith('_auto/') else tag
    if with_hash_tag:
        return f'#{tag}'
    else:
        return tag



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/14_markdown.obsidian.personal.note_type.ipynb.

# %% auto 0
__all__ = ['PersonalNoteTypeEnum', 'type_of_note', 'note_is_of_type', 'NoteTypeError', 'assert_note_is_of_type']

# %% ../../../../nbs/14_markdown.obsidian.personal.note_type.ipynb 3
from deprecated import deprecated
from enum import Enum, auto
import os
from pathlib import Path
from typing import Optional, Union
import warnings

from trouver.helper import (
    file_existence_test
)
from trouver.markdown.obsidian.vault import (
    note_path_by_name, note_name_unique, VaultNote, NoteDoesNotExistError
)

# %% ../../../../nbs/14_markdown.obsidian.personal.note_type.ipynb 6
# TODO: Make these methods private.
def _is_index_note(vault_note: VaultNote) -> bool:
    """Returns `True` if the markdown file is determined to
    be an index note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function only looks at the file name, so it may return
    incorrect results if the note is named incorrectly.
    """
    # file = file_existence_test(file, vault)
    return vault_note.name.startswith('_index')


def _is_index_of_notation_note(vault_note: VaultNote) -> bool:
    """Returns `True` if the markdown file is determined to
    be an index of notation note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function only looks at the file name, so it may return
    incorrect results if the note is named incorrectly.
    """
    return vault_note.name.startswith('_notation')


def _is_standard_information_note(vault_note: VaultNote) -> bool:
    """Returns `True` if the markdown file is determined to
    be a standard information note.

    This function looks for some headings to verify that the note
    is a standard information note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function may return incorrect results, especially
    false negatives.
    - This function returns True if a few tests pass, but will
    also print out a warning with the the note name if not enough
    tests pass.
    """
    # file = file_existence_test(file, vault)
    if (_is_index_of_notation_note(vault_note)
        or _is_template_note(vault_note)):
        return False
    # TODO
    text = vault_note.text()
    tokens = ['# Meta', '## References',
              '## Citations and Footnotes']
    passes = [x in text for x in tokens]
    total_passes = sum(passes)
    if total_passes == 0:
        return False
    elif total_passes <= 2:
        warnings.warn('The following file seems to be a standard information'
                      f' note, but this is ambiguous: {vault_note.name}')
    return True


# TODO: delete
@deprecated
def _is_module_information_note(vault_note: VaultNote) -> bool:
    """Returns `True` if the markdown file is determined to
    be a module information note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function may return incorrect results.
    """
    # TODO figure out which files the card is embedded in
    return 


def _is_notation_note(vault_note: VaultNote) -> bool:
    """Returns `True` if the markdown file is determined to
    be a notation note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function looks at the file name, so it may return
    incorrect results if the note is named incorrectly.
    """
    if (_is_index_of_notation_note(vault_note)
            or _is_standard_information_note(vault_note)):
        return False
    return 'notation' in vault_note.name


def _is_premise_note(vault_note: VaultNote) -> bool:
    """Returns `True` if the markdown file is determined to
    be a premise note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function may return incorrect results.
    """
    # TODO figure out which files the card is embedded in
    # TODO test
    return 'premise' in vault_note.name


def _is_common_terms_note(vault_note: VaultNote) -> bool:
    """Returns `True` if the markdown file is determined to
    be a common terms note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function only looks at the file name, so it may return
    incorrect results if the note is named incorrectly.
    """
    return vault_note.name.startswith('_common_terms')


def _is_template_note(vault_note: VaultNote):
    """Returns `True` if the markdown file is determined to
    be a template note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function only looks at the file name, so it may return
    incorrect results if the note is named incorrectly.
    """
    return vault_note.name.startswith('_template')


# TODO: delete
@deprecated
def _is_tag_note(vault_note: VaultNote):
    return


def _is_notes_note(vault_note: VaultNote):
    # TODO
    """Returns `True` if the markdown file is determined to
    be a notes note.

    A notes note usually is self-written.
    
    """
    return


def _is_mathematician_note(vault_note: VaultNote):
    """Returns `True` if the markdown file is determined to
    be a mathematician note.

    **Parameters**
    - `vault_note` - VaultNote
    
    **Returns**
    - `bool`
    
    **Notes**
    - This function only looks for headers `# Basic` and
    `# Notables`, so it may return incorrect results if the
    note is formatted incorrectly.
    """
    text = vault_note.text()
    tokens = ['# Basic', '# Notables']
    passes = [x in text for x in tokens]
    total_passes = sum(passes)
    return total_passes != 0


def _is_equation_note(vault_note: VaultNote):
    # TODO
    return


def _is_reference_note(vault_note: VaultNote):
    return vault_note.name.startswith('_reference')


def _is_glossary_note(vault_note: VaultNote):
    return vault_note.name.startswith('_glossary')

# %% ../../../../nbs/14_markdown.obsidian.personal.note_type.ipynb 8
class PersonalNoteTypeEnum(Enum):
    """
    An Enum class for note types in my Obsidian math vault
    
    """
    INDEX_NOTE = auto()
    STANDARD_INFORMATION_NOTE = auto()
    MODULE_INFORMATION_NOTE = auto()
    INDEX_OF_NOTATION_NOTE = auto()
    NOTATION_NOTE = auto()
    PREMISE_NOTE = auto()
    COMMON_TERMS_NOTE = auto()
    TEMPLATE_NOTE = auto()
    TAG_NOTE = auto()
    NOTES_NOTE = auto()
    MATHEMATICIAN_NOTE = auto()
    EQUATION_NOTE = auto()
    REFERENCE_NOTE = auto()
    GLOSSARY_NOTE = auto()
    
    @classmethod
    def type_function(cls, note_type: 'PersonalNoteTypeEnum'):
        return cls._type_method_dict[note_type]


PersonalNoteTypeEnum._type_method_dict = {
    PersonalNoteTypeEnum.INDEX_NOTE: _is_index_note,
    PersonalNoteTypeEnum.STANDARD_INFORMATION_NOTE: _is_standard_information_note,
    PersonalNoteTypeEnum.MODULE_INFORMATION_NOTE: _is_module_information_note,
    PersonalNoteTypeEnum.INDEX_OF_NOTATION_NOTE: _is_index_of_notation_note,
    PersonalNoteTypeEnum.NOTATION_NOTE: _is_notation_note,
    PersonalNoteTypeEnum.PREMISE_NOTE: _is_premise_note,
    PersonalNoteTypeEnum.COMMON_TERMS_NOTE: _is_common_terms_note,
    PersonalNoteTypeEnum.TEMPLATE_NOTE: _is_template_note,
    PersonalNoteTypeEnum.TAG_NOTE: _is_tag_note,
    PersonalNoteTypeEnum.NOTES_NOTE: _is_notes_note,
    PersonalNoteTypeEnum.MATHEMATICIAN_NOTE: _is_mathematician_note,
    PersonalNoteTypeEnum.EQUATION_NOTE: _is_equation_note,
    PersonalNoteTypeEnum.REFERENCE_NOTE: _is_reference_note,
    PersonalNoteTypeEnum.GLOSSARY_NOTE: _is_glossary_note
}

# %% ../../../../nbs/14_markdown.obsidian.personal.note_type.ipynb 11
def type_of_note(
        vault_note: VaultNote
        ) -> PersonalNoteTypeEnum:
    """Returns the type of the specified note.

    Assumes that `vault_note` represents an existing note.

    **Notes**
    - This function may return incorrect or unintended results,
    depending on the formatting or existence of the file.
    
    **Raises**
    - NoteDoesNoteExistError
        - If `vault_note` does not exist.
    """
    if not vault_note.exists():
        raise NoteDoesNotExistError(vault_note.name)
    for note_type in PersonalNoteTypeEnum:
        type_function = PersonalNoteTypeEnum.type_function(note_type)
        if type_function(vault_note):
            return note_type

# %% ../../../../nbs/14_markdown.obsidian.personal.note_type.ipynb 12
def note_is_of_type(
        vault_note: VaultNote,
        note_type: Optional[PersonalNoteTypeEnum] # The type of note. If `None`, then any type of note; in this case, returns whether or not the note exists.
        ) -> bool:
    """Returns `True` if the Markdown vault note exists and is determined to
    be of the specified type.

    Assumes that `vault_note` represents an existing note.
    
    **Notes**
    - This function may return incorrect results, depending on the 
    formatting of the file.

    **Raises**
    - NoteDoesNoteExistError
        - If `vault_note` does not exist.
    """
    if not vault_note.exists():
        return False
    return type_of_note(vault_note) == note_type if note_type else True

# %% ../../../../nbs/14_markdown.obsidian.personal.note_type.ipynb 18
class NoteTypeError(ValueError):
    """Exception raised when the type of note is not of the expected type.
    
    **Attributes**
    - `note` - VaultNote
    - `expected_note_type` - PersonalNoteTypeEnum
    """
    def __init__(
            self,
            note: VaultNote,
            expected_note_type: PersonalNoteTypeEnum):
        self.note = note
        self.expected_note_type = expected_note_type
        super().__init__(
            f"Expected a note of type {expected_note_type}, "
            f"but got a note of a different type instead: {note.name}")


def assert_note_is_of_type(
        note: VaultNote,
        expected_type: PersonalNoteTypeEnum):
    """Raises a `NoteTypeError` if the note is not of the
    expected type.
    
    **Raises**
    - NoteTypeError
        - If the type of `note` is not `expected_type`.
    """
    if not note_is_of_type(note, expected_type):
        raise NoteTypeError(note, expected_type)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/11_markdown.obisidian.personal.authors.ipynb.

# %% auto 0
__all__ = ['just_last_name', 'find_author_file']

# %% ../../../../nbs/11_markdown.obisidian.personal.authors.ipynb 2
import os 
from os import PathLike
from pathlib import Path
from typing import Union

# %% ../../../../nbs/11_markdown.obisidian.personal.authors.ipynb 4
def just_last_name(author_name) -> bool:
    """Return `True` if the author's name is formatted to only
    have the last name.
    """
    return '_' not in author_name

# %% ../../../../nbs/11_markdown.obisidian.personal.authors.ipynb 5
# TODO: This method seems redundant because the `all_note_paths_by_name` and `note_name_from_path` functions in `
def find_author_file(
        vault: PathLike,
        author_name: str, # Either just the last name, e.g. `"serre"` or formatted as `"{family_name}_{given_name}_{middle or other initial}"`, e.g.  `"little_john_b"`, `"lockhart_paul"`. In the former case, the mathematician's last name must be unique in the vault. In the latter case, the str must identically match the mathematician's file name in the vault.
        author_folder: str = 'mathematicians' # The directory where the author files are found in, relative to `vault`.
        ) -> Union[PathLike, None]: # Relative to `vault / author_folder`. `None` if no such author file exists.
    """Return path to the file for the specified author.

    **Raises**
    - AssertionError
        - If `author_folder` does not exist in `vault` or 
        if the requirements of `author_name` are not met.
    """
    author_location = Path(vault) / author_folder
    assert os.path.exists(author_location),\
        f"The following path does not exist:{author_location}"
    if just_last_name(author_name):
        notes_of_name = author_location.glob(f'**/{author_name}_*.md')
    else:
        notes_of_name = author_location.glob(f'**/{author_name}.md')
    if notes_of_name:
        assert len(notes_of_name) <= 1,\
            f'Author file is not unique: {author_name}'
        return notes_of_name[0]
    else:
        return None




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/19_markdown.obsidian.personal.note_processing.ipynb.

# %% auto 0
__all__ = ['remove_double_asterisks_in_markdown_file', 'process_standard_information_note']

# %% ../../../../nbs/19_markdown.obsidian.personal.note_processing.ipynb 1
from deprecated import deprecated
from os import PathLike
from pathlib import Path, PureWindowsPath, WindowsPath
import re
from typing import Optional, Union

from ...markdown.file import MarkdownFile, MarkdownLineEnum, replace_embedded_links_with_text
from ..footnotes import remove_footnote_mentions_in_markdown_text
from ..links import EMBEDDED_PATTERN, remove_links_from_text
from ..vault import VaultNote

# %% ../../../../nbs/19_markdown.obsidian.personal.note_processing.ipynb 7
def remove_double_asterisks_in_markdown_file(
        markdown_file: MarkdownFile 
        ) -> None: 
    # TODO Don't remove asterisks in math mode.
    """Remove double asterisks in MarkdownFile object.
    
    The author of `trouver` chose to implement this function outside of the 
    `MarkdownFile` class because its use seems specific - this function
    is mostly intended to remove double asterisks marking definitions and
    notations introduced in LaTeX text.
    """
    for part in markdown_file.parts:
        part['line'] = part['line'].replace('**', '')

# %% ../../../../nbs/19_markdown.obsidian.personal.note_processing.ipynb 9
def process_standard_information_note(
        markdown_file: Union[MarkdownFile, str],
        vault: PathLike,
        remove_frontmatter_meta: bool = True, # If `True`, removes the frontmatter meta. Defaults to `True`
        remove_see_also_section: bool = True, # If `True`, removes the `# See also` section. Defaults to `True`.
        remove_meta_section: bool = True, # If `True`, remove the `# Meta` section. Defaults to `True`.
        remove_references_section: bool = True, # If `True`, removes the `## References` section. Defaults to `True`.
        remove_double_asterisks: bool = True, # If `True`, removes double asterisks. Defaults to `True`.
        remove_links: bool = True, # If `True`, removes nonembedded links and replaces them with their display text. Defaults to `True`.
        remove_in_line_tags: bool = True, # If `True`, removes in-line tags (the lines that start with a tag).  Defaults to `True`.
        remove_footnotes_to_embedded: bool = True, # If `True`, removes footnotes to embedded notes. Defaults to `True`.
        remove_headers: bool = True, # If `True`, removes headers. Defaults to `True`.
        remove_citation_footnotes: bool = True, # If `True`, removes the citation footnote. Defaults to `True`.
        replace_embedded_links_with_content: bool = True, # If `True`, replaces embedded links with their content.  Defaults to `True`.
        merge_display_math_mode: bool = True, # If `True`, merge each group of display math mode latex lines into single lines.  Defaults to `True`.
        merge_display_math_mode_into_text: Optional[str] = None, # If not `None`, merge each group of display math mode latex lines into single lines and merge those groups into the text that precedes them with the specified str. Defaults to `None`.  The blank character ` ` and the new-line character `\n` are recommended as arguments.
        no_double_blank_lines: bool = True # If `True`, removes escape characters `'\n'` to make it so that there are no double blank lines. Defaults to `True`.
        ) -> MarkdownFile: # If `markdown_file` is a `MarkdownFile` object, then the output is `markdown_file` itself (not a copy) with modifications. If `markdown_file` is a `str`, then the output is a `MarkdownFile` object with the modifications.
    """Process/modify a str/MarkdownFile of a standard information note.
        
    TODO: implement remove_citation_footnote properly.
    
    """
    if isinstance(markdown_file, str):
        markdown_file = MarkdownFile.from_string(markdown_file)
    if remove_frontmatter_meta:
        markdown_file.remove_metadata()
    # if not remove_citation_footnote:
    #     return
    if remove_footnotes_to_embedded:
        markdown_file.remove_footnotes_to_embedded_links()
    if replace_embedded_links_with_content:
        markdown_file.replace_embedded_links_with_text(vault)
    if remove_see_also_section:
        markdown_file.remove_section('See Also')
    if remove_meta_section:
        markdown_file.remove_section('Meta')
    if remove_references_section:
        markdown_file.remove_section('References')
    if remove_double_asterisks:
        remove_double_asterisks_in_markdown_file(markdown_file)
    if remove_links: 
        markdown_file.replace_links_with_display_text()
    if remove_in_line_tags:
        markdown_file.remove_in_line_tags()
    if remove_headers:
        markdown_file.remove_headers()
    if merge_display_math_mode:
        markdown_file.merge_display_math_mode()
    if merge_display_math_mode_into_text:
        markdown_file.merge_display_math_mode_into_preceding_text(
            separator=merge_display_math_mode_into_text)
    if no_double_blank_lines:
        markdown_file.remove_double_blank_lines()
    # print(markdown_file)
    return markdown_file

